from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Dict, Tuple

if TYPE_CHECKING:
    from pyspark.sql import DataFrame as SparkDF

from vectice.models.resource.metadata.column_metadata import (
    BooleanStat,
    Column,
    ColumnCategoryType,
    DateStat,
    NumericalStat,
    Quantiles,
    Size,
    TextStat,
)
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper
from vectice.utils.common_utils import safe_float

_logger = logging.getLogger(__name__)

try:
    from pyspark.sql import SparkSession
    from pyspark.sql.functions import avg, expr

    spark_imported = True
except ImportError:
    spark_imported = False

    from vectice.models.resource.metadata.pyspark_typing import avg, expr

    pass


class SparkDFWrapper(DataFrameWrapper):
    def __init__(self, dataframe: SparkDF):
        if spark_imported is False:
            raise ImportError("Pyspark is not installed.")
        super().__init__(dataframe)
        self.spark = SparkSession.builder.getOrCreate()  # type: ignore @see: https://github.com/microsoft/pylance-release/issues/4577

    def get_size(self) -> Size:
        self.rows: int = self.dataframe.count()
        self.columns_numbers: int = len(self.dataframe.columns)

        return Size(rows=self.rows, columns=self.columns_numbers)

    def capture_column_schema(self) -> list[Column]:
        column_cat: ColumnCategoryType | None = None
        list_schema: list[Column] = []

        for column, column_type in self.dataframe.dtypes:
            column = str(column)
            if column_type in ["tinyint", "smallint", "int", "bigint", "float", "double"] or "decimal" in column_type:
                column_cat = ColumnCategoryType.NUMERICAL
                dtype = str(column_type)
            elif column_type == "date" or column_type == "timestamp":
                column_cat = ColumnCategoryType.DATE
                dtype = str(column_type)
            elif column_type == "boolean":
                column_cat = ColumnCategoryType.BOOLEAN
                dtype = str(column_type)
            elif column_type in ["string", "char", "varchar"]:
                column_cat = ColumnCategoryType.TEXT
                dtype = str(column_type)
            else:
                column_cat = None
                dtype = str(column_type)

            list_schema.append(
                Column(
                    name=column,
                    data_type=dtype if dtype != "object" else "string",
                    stats=None,
                    category_type=column_cat,
                )
            )

        return list_schema

    def capture_column_statistics(
        self,
        list_col_schema: list[Column],
        sample_rows_for_statistics: int,
        maximum_columns_for_statistics: int,
    ) -> list[Column]:
        columns: list[Column] = []
        stat: BooleanStat | NumericalStat | TextStat | DateStat | None = None
        result: Dict[str, BooleanStat | NumericalStat | TextStat | DateStat] = {}

        column_split_by_type = self.__get_list_column_split_by_type__(list_col_schema, maximum_columns_for_statistics)
        result_numerical, result_str, result_bool, result_date = self.__compute_spark_stats__(
            column_split_by_type, sample_rows_for_statistics
        )
        result = {**result_numerical, **result_str, **result_bool, **result_date}
        for idx, col in enumerate(list_col_schema):
            if idx == maximum_columns_for_statistics:
                _logger.warning(
                    f"Statistics are only captured for the first {maximum_columns_for_statistics} columns of your dataframe."
                )
            if col.name in result and idx < maximum_columns_for_statistics:
                stat = result[col.name]
            else:
                stat = None

            col.stats = stat
            columns.append(col)
        return columns

    def __compute_spark_stats__(
        self, column_values: Dict[str, list], sample_rows_for_statistics: int
    ) -> Tuple[Dict[str, NumericalStat], Dict[str, TextStat], Dict[str, BooleanStat], Dict[str, DateStat]]:
        result_numerical: Dict[str, NumericalStat] = {}
        result_str: Dict[str, TextStat] = {}
        result_bool: Dict[str, BooleanStat] = {}
        result_date: Dict[str, DateStat] = {}

        # Sampling the DataFrame if needed
        total_rows = self.dataframe.count()
        if sample_rows_for_statistics > 0 and sample_rows_for_statistics < total_rows:
            fraction = sample_rows_for_statistics / total_rows
            sampled_dataframe = self.dataframe.sample(fraction=fraction, seed=42)
        else:
            sampled_dataframe = self.dataframe
        if len(column_values["numerical"]) > 0:
            result_numerical = self.__compute_numeric_column_statistics__(
                sampled_dataframe.select(column_values["numerical"])
            )
        if len(column_values["string"]) > 0:
            result_str = self.__compute_string_column_statistics__(sampled_dataframe.select(column_values["string"]))

        if len(column_values["boolean"]) > 0:
            result_bool = self.__compute_boolean_column_statistics__(sampled_dataframe.select(column_values["boolean"]))

        if len(column_values["date"]) > 0:
            result_date = self.__compute_date_column_statistics__(sampled_dataframe.select(column_values["date"]))
        return result_numerical, result_str, result_bool, result_date

    def __get_list_column_split_by_type__(
        self, list_col_schema: list[Column], maximum_columns_for_statistics: int
    ) -> Dict[str, list]:
        column_split_by_type: Dict[str, list] = {"numerical": [], "string": [], "boolean": [], "date": []}

        for idx, col in enumerate(list_col_schema):
            if idx >= maximum_columns_for_statistics:
                break
            column_name = col.name
            if col.category_type == ColumnCategoryType.NUMERICAL:
                column_split_by_type["numerical"].append(column_name)
            elif col.category_type == ColumnCategoryType.DATE:
                column_split_by_type["date"].append(column_name)
            elif col.category_type == ColumnCategoryType.BOOLEAN:
                column_split_by_type["boolean"].append(column_name)
            elif col.category_type == ColumnCategoryType.TEXT:
                column_split_by_type["string"].append(column_name)
        return column_split_by_type

    def __compute_boolean_column_statistics__(self, dataframe: SparkDF) -> Dict[str, BooleanStat]:
        columns = dataframe.columns
        result_stats: Dict[str, BooleanStat] = {}

        agg_exprs = [
            expr("COUNT(*)").alias("count"),
            *[expr(f"SUM(CASE WHEN `{col}` = True THEN 1 ELSE 0 END)").alias(f"{col}_true") for col in columns],
            *[expr(f"SUM(CASE WHEN `{col}` IS NULL THEN 1 ELSE 0 END)").alias(f"{col}_missing") for col in columns],
            *[expr(f"SUM(CASE WHEN `{col}` = False THEN 1 ELSE 0 END)").alias(f"{col}_false") for col in columns],
        ]

        agg_results = dataframe.agg(*agg_exprs).collect()[0]

        for col in columns:
            count = agg_results["count"]
            true_count = agg_results[f"{col}_true"]
            false_count = agg_results[f"{col}_false"]
            missing_count = agg_results[f"{col}_missing"]

            true_percentage = true_count / count if count > 0 else 0.0
            false_percentage = false_count / count if count > 0 else 0.0
            missing_percentage = missing_count / count if count > 0 and missing_count is not None else 0.0

            result_stats[col] = BooleanStat(
                true=safe_float(true_percentage), false=safe_float(false_percentage), missing=float(missing_percentage)
            )

        return result_stats

    def __compute_numeric_column_statistics__(self, dataframe: SparkDF) -> Dict[str, NumericalStat]:
        columns = dataframe.columns
        result_stats: Dict[str, NumericalStat] = {}

        agg_exprs = [
            expr("COUNT(*)").alias("count"),
            *[expr(f"SUM(CASE WHEN `{col}` IS NULL THEN 1 ELSE 0 END)").alias(f"{col}_missing") for col in columns],
            *[expr(f"AVG(`{col}`)").alias(f"{col}_mean") for col in columns],
            *[expr(f"STDDEV(`{col}`)").alias(f"{col}_stddev") for col in columns],
            *[expr(f"MIN(`{col}`)").alias(f"{col}_min") for col in columns],
            *[expr(f"APPROX_PERCENTILE(`{col}`, 0.25)").alias(f"{col}_q25") for col in columns],
            *[expr(f"APPROX_PERCENTILE(`{col}`, 0.5)").alias(f"{col}_q50") for col in columns],
            *[expr(f"APPROX_PERCENTILE(`{col}`, 0.75)").alias(f"{col}_q75") for col in columns],
            *[expr(f"MAX(`{col}`)").alias(f"{col}_max") for col in columns],
        ]

        agg_results = dataframe.agg(*agg_exprs).collect()[0]
        result_stats = {}

        for col in columns:
            count = agg_results["count"]
            missing_count = agg_results[f"{col}_missing"]
            missing_percentage = missing_count / count if count > 0 and missing_count is not None else 0.0

            mean = agg_results[f"{col}_mean"]
            std_deviation = agg_results[f"{col}_stddev"]
            min_value = agg_results[f"{col}_min"]
            q25 = agg_results[f"{col}_q25"]
            q50 = agg_results[f"{col}_q50"]
            q75 = agg_results[f"{col}_q75"]
            max_value = agg_results[f"{col}_max"]

            result_stats[col] = NumericalStat(
                mean=safe_float(mean),
                std_deviation=safe_float(std_deviation),
                quantiles=Quantiles(
                    q_min=safe_float(min_value),
                    q25=safe_float(q25),
                    q50=safe_float(q50),
                    q75=safe_float(q75),
                    q_max=safe_float(max_value),
                ),
                missing=float(missing_percentage),
            )

        return result_stats

    def __compute_string_column_statistics__(self, dataframe: SparkDF) -> Dict[str, TextStat]:
        columns = dataframe.columns
        result_stats: Dict[str, TextStat] = {}
        agg_exprs = [
            expr("COUNT(*)").alias("count"),
            *[expr(f"SUM(CASE WHEN `{col}` IS NULL THEN 1 ELSE 0 END)").alias(f"{col}_missing") for col in columns],
            *[expr(f"APPROX_COUNT_DISTINCT(`{col}`)").alias(f"{col}_unique") for col in columns],
        ]

        agg_results = dataframe.agg(*agg_exprs).collect()[0]
        result_stats = {}

        for col_name in columns:
            count = agg_results["count"]
            missing_count = agg_results[f"{col_name}_missing"]
            missing_percentage = missing_count / count if count > 0 and missing_count is not None else 0.0

            unique_count = agg_results[f"{col_name}_unique"]

            result_stats[col_name] = TextStat(
                unique=safe_float(unique_count),
                missing=float(missing_percentage),
                most_commons=[],
            )

        return result_stats

    def __compute_date_column_statistics__(self, dataframe: SparkDF) -> Dict[str, DateStat]:
        columns = dataframe.columns
        result_stats: Dict[str, DateStat] = {}

        agg_exprs = [
            expr("COUNT(*)").alias("count"),
            *[expr(f"SUM(CASE WHEN `{col}` IS NULL THEN 1 ELSE 0 END)").alias(f"{col}_missing") for col in columns],
            *[expr(f"MIN(`{col}`)").alias(f"{col}_min") for col in columns],
            *[expr(f"MAX(`{col}`)").alias(f"{col}_max") for col in columns],
            *[avg(expr(f"UNIX_TIMESTAMP(`{col}`, 'yyyy-MM-dd')")).alias(f"{col}_mean") for col in columns],
            *[
                expr(f"APPROX_PERCENTILE(UNIX_TIMESTAMP(`{col}`, 'yyyy-MM-dd'), 0.25, 1000)").alias(f"{col}_median")
                for col in columns
            ],
        ]

        agg_results = dataframe.agg(*agg_exprs).collect()[0]

        for col in columns:
            count = agg_results["count"]
            missing_count = agg_results[f"{col}_missing"]
            missing_percentage = missing_count / count if count > 0 and missing_count is not None else 0.0

            min_date = agg_results[f"{col}_min"]
            max_date = agg_results[f"{col}_max"]
            mean_ts = agg_results[f"{col}_mean"]
            median_ts = agg_results[f"{col}_median"]

            mean_date = (
                datetime.fromtimestamp(int(agg_results[f"{col}_mean"]), timezone.utc) if mean_ts is not None else None
            )
            median_date = (
                datetime.fromtimestamp(int(agg_results[f"{col}_median"]), timezone.utc)
                if median_ts is not None
                else None
            )

            result_stats[col] = DateStat(
                missing=float(missing_percentage),
                minimum=str(min_date),
                mean=str(mean_date),
                median=str(median_date),
                maximum=str(max_date),
            )

        return result_stats



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from vectice.models import AdditionalInfo, Metric, Property
from vectice.models.dataset import TDerivedFrom
from vectice.models.model import Model


class ModelExpTracker(ABC):
    """Base class for Experiment Tracker Wrapper like Mlflow.

    The ModelExpTracker subclasses are used to wrap an experiment tracker model into a Vectice model.
    The `as_vectice_model` method combines the retrieved metrics, properties, attachments, name,
    and derived_from information to create a Vectice Model object representing the
    experiment tracker model.

    Args:
        run: The ID of the experiment run.
        client: The client object used for interacting with the experiment tracking system.
        _derived_from (Optional): A list of derived models from which the
            current model is derived.

    Methods to be implemented:
        get_metrics: Abstract method to retrieve metrics associated with the model.
        get_properties: Abstract method to retrieve properties associated with the model.
        get_attachments: Abstract method to retrieve attachments associated with the model.
        get_name: Abstract method to retrieve the name of the model.

    """

    def __init__(
        self,
        run_id: str,
        client: Any,
        derived_from: list[TDerivedFrom] | None = None,
    ):
        self.run: str = run_id
        self.client = client
        self._derived_from: list[TDerivedFrom] | None = derived_from
        self.name = None

    @abstractmethod
    def get_metrics(self) -> dict[str, int | float] | list[Metric] | Metric | None:
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, str | int] | list[Property] | Property | None:
        pass

    @abstractmethod
    def get_attachments(self) -> list[str] | None:
        pass

    @abstractmethod
    def get_name(self) -> str | None:
        pass

    @abstractmethod
    def get_additional_info(self) -> dict[str, str] | AdditionalInfo | None:
        pass

    def as_vectice_model(self) -> Model:
        metrics = self.get_metrics()
        properties = self.get_properties()
        attachments = self.get_attachments()
        name = self.get_name()
        additional_info = self.get_additional_info()
        model = Model(
            name=name,
            library=None,
            technique=None,
            metrics=metrics,
            properties=properties,
            attachments=attachments,
            derived_from=self._derived_from,
            additional_info=additional_info,
        )
        return model



from __future__ import annotations

from typing import Any

from vectice.api.json.json_type import TJSON
from vectice.api.json.phase import PhaseOutput


class RequirementOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if "phase" in self:
            self._phase: PhaseOutput = PhaseOutput(**self["phase"])
        else:
            self._phase = None  # type: ignore

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        return str(self["description"])



from __future__ import annotations

from typing import Any

from vectice import Dataset, DatasetType, FileResource, Table
from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.service_types.giskard_types import GiskardType
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.utils.common_utils import temp_directory

GISKARD_TEMP_DIR = "giskard"


class GiskardScanReportService(PropertyService, MetricService, TechniqueService):
    def __init__(self, key: str, asset: Any, data: dict):
        self._asset = asset
        self._key = key
        self._temp_dir = temp_directory(GISKARD_TEMP_DIR)

        super().__init__(cell_data=data)

    def _get_rag_model_info(self, model: Any, model_meta: Any) -> dict:
        # predictive function
        params = model_meta.__dict__

        def parse_dict_with_objects_flat(input_dict: dict, prefix: str = "", seen: set | None = None):
            if seen is None:
                seen = set()

            result = {}

            def process_item(key: str, value: dict | Any, current_prefix: str):
                new_key = f"{current_prefix}_{key}" if current_prefix else key

                if isinstance(value, dict):
                    result.update(parse_dict_with_objects_flat(value, f"{new_key}", seen))
                elif hasattr(value, "__dict__"):
                    if id(value) not in seen:
                        seen.add(id(value))
                        class_name = value.__class__.__name__
                        result.update(parse_dict_with_objects_flat(value.__dict__, f"{new_key}_{class_name}", seen))
                        seen.remove(id(value))
                else:
                    result[new_key] = value

            for key, value in input_dict.items():
                process_item(key, value, prefix)

            return result

        parsed_dict = parse_dict_with_objects_flat(params)

        return {
            "variable": self._key,
            "model": model,
            "name": parsed_dict["name"],
            "properties": parsed_dict,
        }

    def _parse_issues(self, issues: list[Any]) -> list[Table]:
        # issues are in groups e.g harmfulness, have a level e.g medium and examples in a pd.DF
        parsed_issues = []

        for issue in issues:
            format_issue = {"section": issue.group.name, "section_description": issue.group.description}
            format_issue["level"] = f"level {issue.level.value}:\n{issue.description}"
            format_issue["table"] = Table(name=f"{issue.group.name} Examples", dataframe=issue.examples(n=5))
            parsed_issues.append(format_issue)
        return parsed_issues

    def _get_report_results_as_dataset(self, report_path: str) -> Dataset:
        filename_path = self._temp_dir / "results_dataset.csv"
        filename = filename_path.as_posix()

        # results df
        results_df = self._asset.to_dataframe()
        results_table = Table(name="Results Table", dataframe=results_df)
        results_df.to_csv(filename)
        resource = FileResource(filename, dataframes=results_df, capture_schema_only=False)
        results_df_name = self._asset.model.name or self._asset.dataset.name
        return Dataset(name=f"{results_df_name} Results", type=DatasetType.UNKNOWN, resource=resource, attachments=[results_table, filename, report_path])  # type: ignore[reportArgumentType]

    def get_asset(self) -> dict[str, Any]:
        asset = {
            "variable": self._key,
            "dataset": None,
            "model": None,
            "scan_report": None,
            "issues": [],
            "asset_type": GiskardType.SCAN,
        }
        # get report html
        report_path = self._temp_dir / "giskard_scan_report.html"
        report_name = report_path.as_posix()
        asset["scan_report"] = report_name
        self._asset.to_html(report_name)
        # input dataset
        asset["input_table"] = Table(name="Question Test Set", dataframe=self._asset.dataset.df)  # type: ignore[reportArgumentType]
        # get the scanner report results as a Dataset with attachments
        asset["results_dataset"] = self._get_report_results_as_dataset(report_name)  # type: ignore[reportArgumentType]
        # parse issues
        asset["issues"] = self._parse_issues(self._asset.issues)

        return asset



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vectice.models.iteration import Iteration
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation


def read_nodejs_date(date_as_string: str | None) -> datetime | None:
    if date_as_string is None:
        return None
    return datetime.strptime(date_as_string, "%Y-%m-%dT%H:%M:%S.%f%z")


def get_asset_type(asset: DatasetVersionRepresentation | ModelVersionRepresentation | Iteration):
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation

    if isinstance(asset, DatasetVersionRepresentation):
        return "DATASET_VERSION"

    if isinstance(asset, ModelVersionRepresentation):
        return "MODEL_VERSION"

    return "ITERATION"


def get_asset_type_url(asset: DatasetVersionRepresentation | ModelVersionRepresentation | Iteration):
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation

    if isinstance(asset, DatasetVersionRepresentation):
        return "datasetversion"

    if isinstance(asset, ModelVersionRepresentation):
        return "modelversion"

    return "iteration"



from __future__ import annotations

import csv
import functools
import logging
import os
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, Literal

from gql import Client as GQLClient
from gql.transport.requests import RequestsHTTPTransport

from vectice.__version__ import __version__
from vectice.api._auth import Auth
from vectice.api._utils import get_asset_type, read_nodejs_date
from vectice.api.attachment import AttachmentApi
from vectice.api.columns_description import ColumnsDescriptionApi
from vectice.api.compatibility import CompatibilityApi
from vectice.api.gql_attachment import GqlAttachmentApi
from vectice.api.gql_dataset import GqlDatasetApi
from vectice.api.gql_feature_flag import GqlFeatureFlagApi
from vectice.api.gql_issue import GqlIssueApi
from vectice.api.gql_lineage import GqlLineageApi
from vectice.api.gql_metric import GqlMetricApi
from vectice.api.gql_model import GqlModelApi
from vectice.api.gql_organization import GqlOrganizationApi
from vectice.api.gql_phase import GqlPhaseApi
from vectice.api.gql_property import GqlPropertyApi
from vectice.api.gql_report import GqlReportApi
from vectice.api.gql_review import GqlReviewApi
from vectice.api.gql_user_workspace_api import UserAndDefaultWorkspaceApi
from vectice.api.http_error_handlers import MissingReferenceError, VecticeException
from vectice.api.iteration import IterationApi
from vectice.api.json import (
    ArtifactName,
    ModelRegisterInput,
    ModelRegisterOutput,
    ModelVersionOutput,
    ModelVersionStatus,
    PagedResponse,
    PropertyInput,
    RequirementOutput,
)
from vectice.api.json.attachment import AttachmentTypeEnum
from vectice.api.json.code_version import CodeVersion
from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_resource_representation import DatasetResourceRepresentationOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.issue import IssueOutput
from vectice.api.json.iteration import (
    IterationContextInput,
    IterationStatus,
    IterationUpdateInput,
    RetrieveIterationOutput,
)
from vectice.api.json.metric import MetricInput
from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput, ModelVersionUpdateInput
from vectice.api.json.organization_config import OrgConfigOutput
from vectice.api.json.report import ReportOutput
from vectice.api.json.review import ReviewOutput
from vectice.api.project import ProjectApi
from vectice.api.project_template import ProjectTemplateApi
from vectice.api.version import VersionApi
from vectice.api.workspace import WorkspaceApi
from vectice.models.attachment_container import FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE
from vectice.models.dataset import Dataset
from vectice.models.iteration import Iteration
from vectice.models.model import Model
from vectice.models.phase import Phase
from vectice.models.representation.attachment import Attachment
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.resource.metadata.base import DatasetSourceType, DatasetSourceUsage, MetadataSettings
from vectice.models.table import Table
from vectice.types.iteration import IterationInput
from vectice.types.phase import PhaseInput
from vectice.types.project import ProjectInput
from vectice.types.version import TVersion
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS
from vectice.utils.vectice_ids_regex import WORKSPACE_VID_REG

if TYPE_CHECKING:
    from io import BytesIO, IOBase

    from requests import Response

    from vectice.api.json import ProjectOutput, ProjectTemplateOutput, WorkspaceOutput
    from vectice.api.json.compatibility import CompatibilityOutput
    from vectice.api.json.iteration import (
        IterationOutput,
        IterationStepArtifact,
        IterationStepArtifactInput,
    )
    from vectice.api.json.phase import PhaseOutput

_logger = logging.getLogger(__name__)


DISABLED_FEATURE_FLAG_MESSAGE = (
    "This '{}' feature is not enabled. Please contact your sales representative for Beta program access."
)


class Client:
    """Low level Vectice API client."""

    _instance: Client | None = None

    def __init__(
        self,
        token: str,
        api_endpoint: str,
    ):
        self.auth = Auth(api_endpoint=api_endpoint, api_token=token)
        transport = RequestsHTTPTransport(url=self.auth.api_base_url + "/graphql", verify=self.auth.verify_certificate)
        logging.getLogger("gql.transport.requests").setLevel("WARNING")
        self._gql_client = GQLClient(transport=transport)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._org_config = self._get_org_config()

    @staticmethod
    def get_instance(
        token: str | None = None,
        api_endpoint: str | None = None,
    ) -> Client:
        if token or api_endpoint:
            Client._instance = None

        if Client._instance is None:
            if token is None or api_endpoint is None:
                raise VecticeException(
                    "You should login before using this method. See https://api-docs.vectice.com/reference/vectice/connection/#vectice.Connection.connect for more info."
                )

            Client._instance = Client(token, api_endpoint)
        return Client._instance

    @property
    def version_api(self) -> str:
        return __version__

    @property
    def version_backend(self) -> str:
        versions = VersionApi(self._gql_client, self.auth).get_public_config().versions
        for version in versions:
            if version.artifact_name == ArtifactName.BACKEND:
                return version.version
        raise ValueError("No version found for backend.")

    def check_compatibility(self) -> CompatibilityOutput:
        return CompatibilityApi(self.auth).check_version()

    def _get_org_config(self) -> OrgConfigOutput:
        return GqlOrganizationApi(self._gql_client, self.auth).get_organization_config()

    def list_projects(
        self,
        workspace: str,
        size: int = DEFAULT_NUMBER_OF_ITEMS,
    ) -> PagedResponse[ProjectOutput]:
        """List the projects in a workspace.

        Parameters:
            workspace: The workspace id.

        Returns:
            The workspace's projects.
        """
        return ProjectApi(self._gql_client, self.auth).list_projects(workspace, size)

    def get_project(self, project: str, workspace: str | None = None) -> ProjectOutput:
        """Get a project.

        Parameters:
            project: The project name or vectice id.
            workspace: The workspace name or id.

        Returns:
            The project JSON structure.
        """
        if workspace is not None and not re.search(WORKSPACE_VID_REG, workspace):
            workspace = WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace).id
        return ProjectApi(self._gql_client, self.auth).get_project(project, workspace)

    def create_project(self, workspace_id: str, project: ProjectInput) -> ProjectOutput:
        return ProjectApi(self._gql_client, self.auth).create_project(workspace_id, project)

    def get_workspace(self, workspace: str) -> WorkspaceOutput:
        """Get a workspace.

        Parameters:
            workspace: The workspace name or id.

        Returns:
            The workspace JSON structure.
        """
        return WorkspaceApi(self._gql_client, self.auth).get_workspace(workspace)

    def list_templates(
        self,
    ) -> list[ProjectTemplateOutput]:
        """List the project templates in a workspace.

        Returns:
            The workspace's project templates.
        """
        return ProjectTemplateApi(self._gql_client, self.auth).list_templates()

    def list_workspaces(self, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[WorkspaceOutput]:
        """List the workspaces.

        Returns:
            The workspaces.
        """
        return WorkspaceApi(self._gql_client, self.auth).list_workspaces(size)

    def upsert_code_attachments(self, files: list[tuple[str, tuple[str, str]]], code_version_id: int):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            code_version_id: The code version id to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files=files, code_version_id=code_version_id)

    def attach_file_to_lineage(self, lineage_id: int, entity_file_id: int):
        """Attach an attachment to an assets lineage.

        Parameters:
            lineage_id: The lineage id to the file to.
            entity_file_id: The file id to attach

        Returns:
            The JSON structure.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).attach_file_to_lineage(
            lineage_id=lineage_id, entity_file_id=entity_file_id
        )

    def upsert_version_attachments(self, files: list[tuple[str, tuple[str, BinaryIO]]], version: TVersion):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            version: The version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_attachment(files, version)

    def upsert_version_tables(self, tables: list[Table], version: TVersion):
        """Upsert a table.

        Parameters:
            tables: The tables to attach.
            version: The version to attach tables to.

        Returns:
            The JSON structure.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).upsert(
            version.id,
            (
                "MODEL_VERSION"
                if isinstance(version, (ModelVersionOutput, ModelVersionRepresentationOutput))
                else "DATASET_VERSION"
            ),
            tables,
        )

    def upsert_iteration_attachments(
        self,
        files: list[tuple[str, tuple[str, BytesIO | IOBase]]],
        iteration_id: str,
        step_name: str | None = None,
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            iteration_id: The iteration id to attach files to.
            step_name (Optional): The step name to add this attachment as an artifact of the step.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, iteration_id=iteration_id, step_name=step_name)

    def upsert_lineage_attachments(
        self,
        files: list[tuple[str, tuple[str, BytesIO | IOBase]]],
        lineage_id: int,
    ):
        """Create an attachment.

        Parameters:
            files: The paths to the files to attach.
            lineage_id: The lineage id to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).create_attachments(files, lineage_id=lineage_id)

    def create_model_predictor(self, model_type: str, model_content: BytesIO, model_version: ModelVersionOutput):
        """Create a predictor.

        Parameters:
            model_type: The type of model to attach.
            model_content: The binary content of the model.
            model_version: The model version to attach files to.

        Returns:
            The JSON structure.
        """
        return AttachmentApi(self.auth).post_model_predictor(model_type, model_content, model_version)

    def get_version_table(
        self,
        asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration,
        table: str,
    ) -> dict[str, Any]:
        """Get a version table.

        Parameters:
            version: The version to get the table from.

        Returns:
            The version table.
        """
        return GqlAttachmentApi(self._gql_client, self.auth).get_version_table(
            asset.id,
            get_asset_type(asset),
            table,
        )

    def list_attachments(
        self, asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration
    ) -> list[Attachment]:
        """List the attachments of an asset.

        Parameters:
            asset: The asset to list attachments from.

        Returns:
            The attachments of the asset.
        """
        attachments = GqlAttachmentApi(self._gql_client, self.auth).list_attachments(
            asset.id,
            get_asset_type(asset),
        )

        return list(
            map(
                lambda attachment: Attachment(
                    id=attachment.id,
                    name=attachment.name,
                    type="file" if attachment.type == AttachmentTypeEnum.ENTITY_FILE else "table",
                ),
                attachments,
            )
        )

    def download_attachment(
        self,
        asset: ModelVersionRepresentation | DatasetVersionRepresentation | Iteration,
        attachment: str,
        file_path: str,
    ) -> None:
        """Download the attachment of a version.

        Parameters:
            asset: The asset to download the attachment from.
            attachment: The attachment to download.
            file_path: The path to download the attachment to.

        Returns:
            None.
        """
        return AttachmentApi(self.auth).download_attachment(asset, attachment, file_path)

    def get_code_version_attachment(self, code_version_id: int, file_id: int) -> Response:
        """Get the attachment of a code version.

        Parameters:
            code_version_id: The code version id to list attachments from.
            file_id: The file id attached to the code version.

        Returns:
            The file attached to the code version.
        """
        return AttachmentApi(self.auth).get_code_version_attachment(code_version_id=code_version_id, file_id=file_id)

    def list_phases(self, project: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[PhaseOutput]:
        return GqlPhaseApi(self._gql_client, self.auth).list_phases(project, size)

    def get_phase(self, phase: str, project_id: str | None = None) -> PhaseOutput:
        if project_id is None:
            raise MissingReferenceError("project")
        return GqlPhaseApi(self._gql_client, self.auth).get_phase(phase, project_id)

    def create_phase(self, project_id: str, phase: PhaseInput) -> PhaseOutput:
        return GqlPhaseApi(self._gql_client, self.auth).create_phase(project_id, phase)

    def get_full_phase(self, phase: str) -> PhaseOutput:
        return GqlPhaseApi(self._gql_client, self.auth).get_phase(phase=phase, full=True)

    def list_sections(self, iteration_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).list_sections(iteration_id)

    def list_reviews(self, phase: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[ReviewOutput]:
        return GqlReviewApi(self._gql_client, self.auth).list_reviews(phase, size)

    def list_resources_items(
        self, id: str, type: DatasetSourceUsage | None
    ) -> list[DatasetResourceRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_resources_items(id, type)

    def list_columns(self, id: int, type: DatasetSourceType) -> list[Dict[str, Any]]:
        return GqlDatasetApi(self._gql_client, self.auth).get_columns(id, type)

    def list_iterations(
        self,
        phase: str,
        only_mine: bool = False,
        statuses: list[IterationStatus] | None = None,
        size: int = DEFAULT_NUMBER_OF_ITEMS,
    ) -> PagedResponse[IterationOutput]:
        return IterationApi(self._gql_client, self.auth).list_iterations(phase, only_mine, statuses, size)

    def list_iteration_assets(self, iteration: str) -> PagedResponse[IterationStepArtifact]:
        return IterationApi(self._gql_client, self.auth).list_iteration_assets(iteration)

    def list_step_definitions(self, phase: str) -> PagedResponse[RequirementOutput]:
        return GqlPhaseApi(self._gql_client, self.auth).list_step_definitions(phase)

    def get_last_iteration(self, phase_id: str, iteration: IterationInput | None = None) -> RetrieveIterationOutput:
        return IterationApi(self._gql_client, self.auth).get_last_iteration(phase_id, iteration)

    def get_active_iteration_or_create(self, phase_id: str) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_active_iteration_or_create(phase_id)

    def get_iteration_by_id(self, iteration_id: str, full: bool = False) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_id(iteration_id, full)

    def get_iteration_by_index(self, phase_id: str, index: int) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).get_iteration_by_index(phase_id, index)

    def create_iteration(self, phase_id: str, iteration: IterationInput | None = None) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).create_iteration(phase_id, iteration)

    def update_iteration(self, iteration_id: str, iteration: IterationUpdateInput) -> IterationOutput:
        return IterationApi(self._gql_client, self.auth).update_iteration(iteration, iteration_id)

    def delete_iteration(self, iteration_id: str) -> None:
        IterationApi(self._gql_client, self.auth).delete_iteration(iteration_id)

    def add_iteration_artifacts(
        self, iteration_id: str, artifacts: list[IterationStepArtifactInput], section: str | None = None
    ) -> list[IterationStepArtifact]:
        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        return IterationApi(self._gql_client, self.auth).add_iteration_artifacts(
            artifacts=artifacts, iteration_context=iteration_context
        )

    def remove_iteration_assets(self, iteration_id: str, section: str | None = None, all: bool = False):
        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        return IterationApi(self._gql_client, self.auth).remove_iteration_assets(
            iteration_context=iteration_context, all=all
        )

    def remove_assets_from_iteration(self, iteration_id: str, artifacts_id_list: list[int]):
        return IterationApi(self._gql_client, self.auth).remove_assets_from_iteration(iteration_id, artifacts_id_list)

    def save_autolog_cells(self, iteration_id: str, cells: list[dict[str, str]], prefix: str | None = None):
        return IterationApi(self._gql_client, self.auth).save_autolog_cells(iteration_id, cells, prefix)

    def organize_with_ai(self, iteration_id: str):
        return IterationApi(self._gql_client, self.auth).organize_with_ai(iteration_id)

    def register_dataset_from_source(
        self,
        dataset: Dataset,
        iteration_id: str,
        code_version: CodeVersion | None = None,
        section: str | None = None,
    ) -> DatasetRegisterOutput:
        name = self.get_dataset_name(dataset)
        derived_from = self.get_derived_from(dataset)
        metadata_asdict, is_updating_columns = self._transform_resource_to_metadata_dict(dataset)
        properties, _ = self.get_properties_and_metrics(dataset)

        dataset_register_input = DatasetRegisterInput(
            name=name,
            type=dataset.type.value,
            datasetSources=metadata_asdict,
            datasetInputs=[s for s in derived_from if s.startswith("DTV-")] if derived_from else None,
            modelInputs=[s for s in derived_from if s.startswith("MDV-")] if derived_from else None,
            codeVersion=code_version.asdict() if code_version is not None else None,
            properties=properties,
        )

        iteration_context = IterationContextInput(iterationId=iteration_id, stepName=section)
        dataset_register_output = self.register_dataset(dataset_register_input, iteration_context=iteration_context)
        dtv_id = dataset_register_output["datasetVersion"]["vecticeId"]

        if is_updating_columns:
            self.warn_if_dataset_version_columns_are_missing_description(dtv_id)

        dataset.latest_version_id = dtv_id
        return dataset_register_output

    def warn_if_dataset_version_columns_are_missing_description(self, dataset_version_id: str):
        has_columns_without_description = GqlDatasetApi(
            self._gql_client, self.auth
        ).get_dataset_version_has_columns_without_description(dataset_version_id)
        if has_columns_without_description is True:
            _logger.warning(
                "Some column descriptions are missing. Consider adding descriptions for improved data clarity."
            )

    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset(id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_version(id)

    def update_dataset_version(self, id: str, columns_description: list[dict[str, str]]):
        return GqlDatasetApi(self._gql_client, self.auth).update_dataset_version(id, columns_description)

    def update_columns_description_via_csv(self, id: str, file_path: str):
        if not os.path.exists(file_path):
            raise ValueError(FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE % file_path)
        curr_file = ("file", (file_path, open(file_path, "rb")))
        ColumnsDescriptionApi(self.auth).post_columns_description(id, curr_file)
        curr_file[1][1].close()

    def get_model(self, id: str) -> ModelRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model(id)

    def get_dataset_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_list(project_id, size)

    def get_dataset_version_list(
        self, id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetVersionRepresentationOutput]:
        return GqlDatasetApi(self._gql_client, self.auth).get_dataset_version_list(id, size)

    def get_model_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ModelRepresentationOutput]:
        return GqlModelApi(self._gql_client, self.auth).get_model_list(project_id, size)

    def get_model_version_list(
        self, id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ModelVersionRepresentationOutput]:
        return GqlModelApi(self._gql_client, self.auth).get_model_version_list(id, size)

    def get_model_version(self, id: str) -> ModelVersionRepresentationOutput:
        return GqlModelApi(self._gql_client, self.auth).get_model_version(id)

    def get_model_version_approval_history(self, id: str) -> dict[str, datetime | None]:
        response = GqlModelApi(self._gql_client, self.auth).get_model_version_approval_history(id)

        def _transform_dict(acc: dict[str, datetime | None], item: tuple[str, str | None]):
            key, value = item
            camel_case_key = re.sub(r"(?<!^)(?=[A-Z])", "_", key).lower()
            return {**acc, camel_case_key: read_nodejs_date(value) if value else None}

        return functools.reduce(_transform_dict, list(response.items()), {})

    def get_lineage_inputs(
        self, id: str, type: Literal["mdv"] | Literal["dtv"]
    ) -> list[DatasetVersionRepresentationOutput]:
        return GqlLineageApi(self._gql_client, self.auth).get_lineage_inputs(id, type)

    def get_lineage_children(
        self, id: str, type: Literal["mdv"] | Literal["dtv"]
    ) -> tuple[list[ModelVersionRepresentationOutput], list[DatasetVersionRepresentationOutput]]:
        return GqlLineageApi(self._gql_client, self.auth).get_lineage_children(id, type)

    def get_reports(
        self, id: str, type: Literal["mdv"] | Literal["prj"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ReportOutput]:
        return GqlReportApi(self._gql_client, self.auth).get_reports(id, type, size)

    def get_issues(
        self, id: str, type: Literal["mdv"] | Literal["prj"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[IssueOutput]:
        return GqlIssueApi(self._gql_client, self.auth).get_issues(id, type, size)

    def update_model(self, model_version_id: str, model: ModelVersionUpdateInput):
        GqlModelApi(self._gql_client, self.auth).update_model(model_version_id, model)

    def has_column(self, id: str, column: str) -> bool:
        return GqlDatasetApi(self._gql_client, self.auth).has_column(id, column)

    @staticmethod
    def get_dataset_name(dataset: Dataset) -> str:
        return (
            f"dataset {datetime.time}"
            if dataset.name is None  # pyright: ignore[reportUnnecessaryComparison]
            else dataset.name
        )

    @staticmethod
    def get_derived_from(obj: Dataset | Model) -> list[str] | None:
        return None if obj.derived_from is None else obj.derived_from  # pyright: ignore[reportUnnecessaryComparison]

    @staticmethod
    def get_properties_and_metrics(
        obj: Dataset | Model,
    ) -> tuple[list[dict[str, Any]] | None, list[dict[str, Any]] | None]:
        properties = [vars(PropertyInput(prop.key, prop.value)) for prop in obj.properties] if obj.properties else None
        metrics = (
            [vars(MetricInput(met.key, met.value)) for met in obj.metrics]
            if (isinstance(obj, Model) and obj.metrics)
            else None
        )
        return properties, metrics

    def _transform_resource_to_metadata_dict(self, dataset: Dataset):
        resources = dataset.resource if isinstance(dataset.resource, tuple) else [dataset.resource]
        metadata_asdict = []
        is_updating_columns = False
        for resource in resources:
            if resource is not None and resource.metadata:
                resource.metadata.set_settings(
                    MetadataSettings(
                        minimum_rows_for_statistics=self._org_config.configuration.df_statistics_row_threshold,
                        sample_rows_for_statistics=self._org_config.configuration.df_statistics_row_sample,
                        maximum_columns_for_statistics=self._org_config.configuration.df_statistics_colmns_threshold,
                    )
                )

                metadata, columns_description = resource.metadata, resource.columns_description
                is_updating_columns = is_updating_columns is False and columns_description is not None
                columns = columns_description or {}
                metadata_dict, type = metadata.asdict(), metadata.type
                if type == DatasetSourceType.DB:
                    for db in metadata_dict["dbs"]:
                        db["columns"] = self._combine_source_column_description(db, columns)
                else:
                    for file in metadata_dict["files"]:
                        file["columns"] = self._combine_source_column_description(file, columns)

                metadata_asdict.append(metadata_dict)
        return metadata_asdict, is_updating_columns

    def _combine_source_column_description(self, source: dict, columns_description: dict[str, str] | str):
        def _transform_path_to_dictionary(path: str):
            def _reduce_csv(acc: dict[str, str], curr: list[str]):
                if len(curr) > 1:
                    return {**acc, curr[0]: curr[1]}
                return acc

            with open(path, mode="r", newline="") as csv_file:
                return functools.reduce(_reduce_csv, csv.reader(csv_file), {})

        if isinstance(columns_description, str):
            columns_description = _transform_path_to_dictionary(columns_description)

        def _map_columns(col: dict):
            col_name = col["name"]
            if col_name in columns_description:
                column_desc = columns_description[col_name]
                return {**col, "description": column_desc}
            return col

        return list(map(_map_columns, source["columns"]))

    def register_dataset(
        self,
        dataset_register_input: DatasetRegisterInput,
        iteration_context: IterationContextInput,
    ) -> DatasetRegisterOutput:
        data: DatasetRegisterOutput = GqlDatasetApi(self._gql_client, self.auth).register_dataset(
            dataset_register_input, iteration_context
        )
        _logger.debug(
            f"Successfully registered Dataset("
            f"name='{dataset_register_input.name}', "
            f"id={data['datasetVersion']['vecticeId']}, "
            f"version='{data['datasetVersion']['name']}', "
            f"type={dataset_register_input.type})."
        )
        return data

    def register_model(
        self,
        model: Model,
        iteration: Iteration,
        phase: Phase,
        code_version: CodeVersion | None = None,
        section: str | None = None,
    ) -> ModelRegisterOutput:
        """Register a model.

        Parameters:
            model: The model to register
            iteration: The iteration
            phase: The phase
            code_version_id: The code version ID
            section: The section

        Returns:
            The registered model.
        """
        properties, metrics = self.get_properties_and_metrics(model)
        derived_from = self.get_derived_from(model)
        model.name = (model.name or f"{phase.name} {iteration.index} model")[:60]
        model_register_input = ModelRegisterInput(
            name=model.name,
            modelType="OTHER",
            status=ModelVersionStatus.EXPERIMENTATION.value,
            datasetInputs=[s for s in derived_from if s.startswith("DTV-")] if derived_from else None,
            modelInputs=[s for s in derived_from if s.startswith("MDV-")] if derived_from else None,
            metrics=metrics,
            properties=properties,
            algorithmName=model.technique,
            framework=model.library,
            codeVersion=code_version.asdict() if code_version is not None else None,
            context=model.additional_info.asdict() if model.additional_info is not None else None,
        )
        iteration_context = IterationContextInput(iterationId=iteration.id, stepName=section)

        model_output = GqlModelApi(self._gql_client, self.auth).register_model(
            model_register_input, iteration_context=iteration_context
        )
        model.latest_version_id = model_output["modelVersion"]["vecticeId"]
        return model_output

    def upsert_properties(self, type: str, id: str, properties: list[dict[str, str | int]]):
        return GqlPropertyApi(self._gql_client, self.auth).upsert(type, id, properties)

    def upsert_metrics(self, type: str, id: str, metrics: list[dict[str, str | float | int]]):
        return GqlMetricApi(self._gql_client, self.auth).upsert(type, id, metrics)

    def get_user_and_default_workspace(self):
        return UserAndDefaultWorkspaceApi(self._gql_client, self.auth).get_user_and_default_workspace()

    def assert_feature_flag_or_raise(self, code: str) -> None:
        enabled = self.is_feature_flag_enabled(code)
        if enabled is False:
            raise VecticeException(DISABLED_FEATURE_FLAG_MESSAGE.format(code))

    def is_feature_flag_enabled(self, code: str) -> bool:
        enabled = GqlFeatureFlagApi(self._gql_client, self.auth).is_feature_flag_enabled(code)
        if enabled is False:
            _logger.info(DISABLED_FEATURE_FLAG_MESSAGE.format(code))
        return enabled

    def is_feature_flag_enabled_without_message(self, code: str) -> bool:
        return GqlFeatureFlagApi(self._gql_client, self.auth).is_feature_flag_enabled(code)



from __future__ import annotations

from functools import reduce
from typing import Type, TypeVar, overload

from vectice.api.json.json_type import TJSON

ItemType = TypeVar("ItemType")


@overload
def json_to_class(json: TJSON, cls: Type[ItemType]) -> ItemType: ...


@overload
def json_to_class(json: list[TJSON], cls: Type[ItemType]) -> list[ItemType]: ...


def json_to_class(json: TJSON | list[TJSON], cls: Type[ItemType]) -> ItemType | list[ItemType]:
    if isinstance(json, list):
        return reduce(lambda acc, curr: [*acc, cls(**curr)], json, [])

    return cls(**json)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from giskard.scanner.scanner import ScanReport  # type: ignore[reportMissingImports]
    from pandas import DataFrame


from giskard import Dataset, Model, scan  # type: ignore[reportMissingImports]
from giskard.scanner.scanner import ScanReport  # type: ignore[reportMissingImports]


def giskard_scan_llm_wrapper(
    context_retrieval_pipeline: Any | None = None,
    model_prediction_function: Callable | None = None,
    custom_model: Model | None = None,
    feature_names: list[str] | None = None,
    model_name: str | None = None,
    model_description: str | None = None,
    question_dataset: DataFrame | None = None,
    dataset_target: str | None = None,
    scan_only: list[str] = ["hallucination"],
    scan_max_issues_per_detector: int = 10,
) -> ScanReport:
    """Wrapper function to configure and execute a model scan using Giskard's scanning capabilities.
    Scans which are executed are harmful content and hallucinations.

    This function initializes a scan configuration by setting up the index, extracting relevant
    data, and configuring a prediction function for the model. If no custom prediction function is
    provided, a default one is created.

    Args:
        index_zip_path (str): The file path to the zipped index containing the preprocessed data.
        index_extract_dir (str): The directory where the index will be extracted.
        qa_testset_path (str): The file path to the QA test set used for model evaluation.
        model_predict_fn (Callable | None, optional): Custom prediction function for the model.
            Defaults to None, which will use a predefined function with simulated injections.
        model_type (str | None, optional): The type of model being scanned (e.g., "LLM", "classification").
            Defaults to None.
        model_name (str | None, optional): A name for the model being scanned. Defaults to None.
        model_description (str | None, optional): A description of the model. Defaults to None.
        feature_names (list[str] | None, optional): List of feature names to be used during the scan.
            Defaults to None.

    Returns:
        ScanReport: A report containing the results of the scan, highlighting issues related to
        the model's responses.

    Example:
        report = giskard_scan_wrapper(
            index_zip_path="/path/to/index.zip",
            index_extract_dir="/path/to/extracted/index",
            qa_testset_path="/path/to/qa_testset.jsonl",
            model_type="LLM",
            model_name="Test Model",
            model_description="A large language model for testing purposes.",
            feature_names=["question"]
        )
    """
    if isinstance(custom_model, Callable):
        # Define a custom Giskard model wrapper for the serialization.
        ### generic custom model
        giskard_model = custom_model(
            model=context_retrieval_pipeline,
            model_type="text_generation",  # type: ignore[reportArgumentType]
            name=model_name,
            description=model_description,
            feature_names=feature_names,
        )
    elif context_retrieval_pipeline and model_prediction_function:
        giskard_model = Model(
            model=context_retrieval_pipeline,
            # assumption for now...
            model_postprocessing_function=model_prediction_function,
            model_type="text_generation",  # type: ignore[reportArgumentType]
            name=model_name,
            description=model_description,
            feature_names=feature_names,
        )
    elif model_prediction_function and (custom_model is None and context_retrieval_pipeline is None):
        giskard_model = Model(
            model=model_prediction_function,
            # assumption for now...
            model_type="text_generation",  # type: ignore[reportArgumentType]
            name=model_name,
            description=model_description,
            feature_names=feature_names,
        )
    else:
        raise ValueError("no model")
    # used to validate the model, done prior to scan.
    giskard_dataset = None
    if question_dataset is not None:
        giskard_dataset = Dataset(question_dataset, dataset_target)

    return scan(giskard_model, giskard_dataset, only=scan_only, max_issues_per_detector=scan_max_issues_per_detector)  # type: ignore[reportCallIssue]



from __future__ import annotations

from vectice.api.json.json_type import TJSON


class OrganizationConfigOutput(TJSON):
    @property
    def df_statistics_row_threshold(self) -> int:
        return int(self["dfStatisticsRowThreshold"])

    @property
    def df_statistics_row_sample(self) -> int:
        return int(self["dfStatisticsSampleRows"])

    @property
    def df_statistics_colmns_threshold(self) -> int:
        return int(self["dfStatisticsMaxColumns"])


class OrgConfigOutput(TJSON):
    @property
    def configuration(self) -> OrganizationConfigOutput:
        return OrganizationConfigOutput(self["organization"]["configuration"])



from __future__ import annotations

from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ExtraMetadata:
    key: str
    value: str | None
    display_name: str | None



from __future__ import annotations

from enum import Enum


class VecticeType(Enum):
    MODEL = "MODEL"
    METRIC = "METRIC"
    DATASET = "DATASET"
    VECTICE_OBJECT = "VECTICE_OBJECT"
    IPYWIDGETS_OUTPUT = "IPYWIDGETS_OUTPUT"



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput

if TYPE_CHECKING:
    from datetime import datetime


class ModelRepresentationOutput(TJSON):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def description(self) -> str | None:
        return str(self["description"]) if self["description"] else None

    @property
    def version(self) -> ModelVersionRepresentationOutput | None:
        if self["lastVersion"] is not None:
            return ModelVersionRepresentationOutput(**self["lastVersion"])
        else:
            return None

    @property
    def project_id(self) -> str | None:
        if "project" in self:
            return str(self["project"]["vecticeId"])
        return None

    @property
    def total_number_of_versions(self) -> int:
        return int(self["versionCount"])

    @property
    def version_stats(self) -> dict[str, int] | None:
        if "versionStats" in self:
            return self["versionStats"]
        return self["versionStats"]



from vectice.autolog.asset_services.service_types.giskard_types import GiskardType
from vectice.autolog.asset_services.service_types.modeva_types import ModevaType
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType

__all__ = [
    "GiskardType",
    "ModevaType",
    "VecticeType",
]



from __future__ import annotations

import logging
import pickle  # nosec
from typing import TYPE_CHECKING, Any, List

from vectice.models.attachment import TAttachment, TFormattedAttachment

if TYPE_CHECKING:
    from mlflow import MlflowClient  # pyright: ignore[reportPrivateImportUsage]

from vectice.models import AdditionalInfo, ExtraInfo, Metric, Property
from vectice.models.dataset import TDerivedFrom, get_derived_from
from vectice.utils.common_utils import check_key_duplicates, format_attachments, format_metrics, format_properties

_logger = logging.getLogger(__name__)


class Model:
    """Represent a wrapped model.

    To document models within an iteration, you need to wrap it as a Vectice model. It involves capturing metadata about the predictor.
    This metadata enables Vectice to provide summaries and insights about the model.
    """

    def __init__(
        self,
        library: str | None = None,
        technique: str | None = None,
        metrics: dict[str, int | float] | list[Metric] | Metric | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        name: str | None = None,
        attachments: TAttachment | None = None,
        predictor: Any = None,
        derived_from: list[TDerivedFrom] | None = None,
        additional_info: dict[str, str] | AdditionalInfo | None = None,
    ):
        """Wrap a model (predictor).

        A Vectice Model is a wrapped predictor that can be logged
        inside a Vectice iteration.

        Parameters:
            library: The library used to generate the model.
            technique: The modeling technique used.
            metrics: A dict for example `{"MSE": 1}`.
            properties: A dict, for example `{"folds": 32}`.
            name: The model name. If None, will be auto-generated by Vectice.
            attachments: The file paths that will be attached to the iteration along with the predictor.
            predictor: The predictor.
            derived_from: List of dataset versions or model versions (or their ids) to link as lineage.
            additional_info: An optional set of key values related to the context in which the model was created. For instance if an experiment tracker was used or a specific framework.
        """
        from vectice import pickle_capture

        self._library = library
        self._technique = technique
        self._name = name or self._generate_name()
        self._metrics = format_metrics(metrics) if metrics else None
        self._properties = format_properties(properties) if properties else None
        self._attachments = format_attachments(attachments) if attachments else None
        self._derived_from = get_derived_from(derived_from)
        self._additional_info = self._format_additional_info(additional_info) if additional_info else None
        self._latest_version_id: str | None = None
        self._predictor = predictor if pickle_capture else None  # nosec

    def __repr__(self):
        return (
            f"Model(name='{self.name or ''}', library='{self.library or ''}', technique='{self.technique or ''}', "
            f"metrics={self.metrics}, properties={self.properties}, attachments={self.attachments}, additional_info={self.additional_info})"
        )

    @property
    def latest_version_id(self) -> str | None:
        """The id of the latest version of this model.

        Returns:
            The id of the latest version of this model.
        """
        return self._latest_version_id

    @latest_version_id.setter
    def latest_version_id(self, value: str) -> None:
        """Set the id of the latest version of this model.

        Parameters:
            value: The id of the latest version of this model.
        """
        self._latest_version_id = value

    @property
    def name(self) -> str | None:
        """The model's name.

        Returns:
            The model's name.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Set the model's name.

        Parameters:
            name: The name of the model.
        """
        self._name = name

    @property
    def predictor(self) -> Any:
        """The model's predictor.

        Returns:
            The model's predictor.
        """
        return self._predictor  # noqa: S301

    @predictor.setter
    def predictor(self, predictor: Any):
        """Set the model's predictor.

        Parameters:
            predictor: The predictor of the model.
        """
        self._predictor = predictor

    @classmethod  # pyright: ignore[reportArgumentType]
    def read_predictor_file(cls, path: str) -> Any:
        with open(path, "rb") as fh:
            return pickle.load(fh)  # noqa: S301

    @property
    def library(self) -> str | None:
        """The name of the library used to generate the model.

        Returns:
            The name of the library used to generate the model.
        """
        return self._library

    @library.setter
    def library(self, library: str):
        """Set the name of the library used to create the model.

        Parameters:
            library: The name of the library used to create the model.
        """
        self._library = library

    @property
    def technique(self) -> str | None:
        """The name of the modeling technique used to learn the model.

        Returns:
            The name of the modeling technique used to learn the model.
        """
        return self._technique

    @technique.setter
    def technique(self, technique: str):
        """Set the name of the modeling technique used to learn the model.

        Parameters:
            technique: The modeling technique used.
        """
        self._technique = technique

    @property
    def metrics(self) -> list[Metric] | None:
        """The model's metrics.

        Returns:
            The model's metrics.
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: dict[str, int | float] | list[Metric] | Metric | None):
        """Set the model's metrics.

        Parameters:
            metrics: The metrics of the model.
        """
        _logger.warning("To save your updated model metrics, you must reassign your model to an iteration.")
        self._metrics = format_metrics(metrics)

    @property
    def properties(self) -> list[Property] | None:
        """The model's properties.

        Returns:
            The model's properties.
        """
        return self._properties

    @properties.setter
    def properties(self, properties: dict[str, str | int] | list[Property] | Property | None):
        """Set the model's properties.

        Parameters:
            properties: The properties of the model.
        """
        _logger.warning("To save your updated model properties, you must reassign your model to an iteration.")
        self._properties = format_properties(properties) if properties else None

    @property
    def attachments(self) -> list[TFormattedAttachment] | None:
        """The file attachment objects or paths associated with the model.

        Returns:
        `List[Union[Table, str]] | None`
            A list of attachment objects, where each attachment is one of:

        - `Table`
            A formatted table object containing structured data
        - `str`
            A string representation of an attachment (e.g., file path)
        - `None`
            If no attachments are present
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: TAttachment):
        """Attach a file or files to the model.

        Parameters:
            attachments: The filename or filenames of the file or set of files to attach to the model.
        """
        self._attachments = format_attachments(attachments)

    @property
    def additional_info(self) -> AdditionalInfo | None:
        """The additional info associated with the model.

        Returns:
            The additional info associated with the model.
        """
        return self._additional_info

    @additional_info.setter
    def additional_info(self, additional_info: dict[str, str] | AdditionalInfo):
        """Attach an additional info to the model.

        Parameters:
            additional_info: The additional information to be passed to the model linked to Experiment Trackers or other context in which the model was created.
        """
        self._additional_info = self._format_additional_info(additional_info)

    @property
    def derived_from(self) -> list[str]:
        """The datasets versions or models versions from which this model is derived.

        Returns:
            The datasets versions or models versions from which this model is derived.
        """
        return self._derived_from

    def _generate_name(self) -> str | None:
        if self.library is not None and self.technique is not None:
            return f"{self.library} {self.technique} model"
        return None

    @staticmethod  # pyright: ignore[reportArgumentType]
    def _remove_incorrect_extra_info(extra_info: list[ExtraInfo]) -> list[ExtraInfo]:
        for info in extra_info:
            if not isinstance(info, ExtraInfo):  # pyright: ignore[reportUnnecessaryIsInstance]
                logging.warning(f"Incorrect extra info '{info}'. Please check extra info type.")
                extra_info.remove(info)
        return extra_info

    def _format_additional_info(self, additional_info: dict[str, str] | AdditionalInfo | None) -> AdditionalInfo | None:
        if additional_info is None:
            return None
        if isinstance(additional_info, AdditionalInfo):
            if isinstance(additional_info.extra_info, list):
                extra_info = self._remove_incorrect_extra_info(additional_info.extra_info)
                additional_info.extra_info = extra_info
                key_list = [info.key for info in additional_info.extra_info]
                check_key_duplicates(key_list)
            return additional_info

        if isinstance(additional_info, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
            extra_info_list: List[ExtraInfo] = []
            for key, value in additional_info.items():
                extra_info_list.append(ExtraInfo(key=key, value=value))
            return AdditionalInfo(url=None, run=None, framework=None, extra_info=extra_info_list)
        else:
            raise ValueError("Please check additional info type.")

    @staticmethod  # pyright: ignore[reportArgumentType]
    def mlflow(
        run_id: str,
        client: MlflowClient,
        url: str | None = None,
        derived_from: list[TDerivedFrom] | None = None,
    ) -> Model:
        """Extract automatically information from an Mlflow run into a Vectice model to be assigned to an iteration.

        ```python
        import mlflow
        from mlflow.client import MlflowClient
        from vectice import Model

        model = Model.mlflow(
                run_id="d479ca87954f4a0abc4da3333b990cb3",
                client=MlflowClient() #also work with client=mlflow
            )
        ```


        Parameters:
            run_id: The run id of the mlflow model to be wrapped.
            client: The mlflow client in order to access the run through get_run() method. WARNING: the client should have the tracking URI where the run is located.
            url: The Mlflow UI URL to access the run info.
            derived_from: List of dataset versions or model versions (or their ids) to link as lineage.
        """
        from vectice.models.model_mlflow import MlflowModel

        model = MlflowModel(run_id, client=client, url=url, derived_from=derived_from).as_vectice_model()
        return model



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark.sql import Column
    from pyspark.sql._typing import ColumnOrName  # pyright: ignore[reportPrivateUsage]


def expr(str: str) -> Column: ...


def avg(col: ColumnOrName) -> Column: ...



from __future__ import annotations

from gql import gql

from vectice.api.gql_api import GqlApi
from vectice.utils.api_utils import PAGINATE_OUTPUT

_RETURNS = PAGINATE_OUTPUT.format(
    """
    key
    value
    __typename
"""
)


class GqlPropertyApi(GqlApi):
    def upsert(self, type: str, id: str, properties: list[dict[str, str | int]]) -> list[dict[str, str | int]]:
        variable_types = (
            "$entityType:EntityPropertyType!,$entityId:VecticeId!,$upsertPropertyList:[UpsertEntityPropertyInput!]!"
        )
        kw = "entityType:$entityType,entityId:$entityId,upsertPropertyList:$upsertPropertyList"
        variables = {"entityType": type, "entityId": id, "upsertPropertyList": properties}
        gql_query = "upsertEntityProperties"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        return self.execute(query_built, variables)[gql_query]["items"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from vectice.api.json.dataset_resource_representation import DatasetResourceRepresentationOutput
from vectice.utils.common_utils import repr_class

if TYPE_CHECKING:
    from vectice.api.client import Client


class ResourceItemRepresentation:
    """Represents the metadata of a Vectice resource.

    An Resource Item Representation shows information about a specific resource item from the Vectice app.
    It makes it easier to get and read this information through the API.

    Attributes:
        path (str): The path of the resource.
        name (str): The name of the resource.
        usage (str): The usage of the resource.
        size (int | None): The size of the resource.
        columns_number (int | None): The columns number of the resource.
        rows_number (int | None): The number of rows of the resource.
        info (dict | None): The extra information of the resource.
    """

    def __init__(self, output: DatasetResourceRepresentationOutput, client: Client):
        self.name = output.name
        self.resource_type = output.usage
        self.path = output.path
        self.size = output.size
        self.columns_number = output.columns_count
        self.rows_number = output.rows_count
        self.info = output.info
        self._type = output.type
        self._client = client
        self._id = output.id

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the ResourceItemRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        return {
            "name": self.name,
            "resource_type": self.resource_type,
            "path": self.path,
            "size": self.size,
            "columns_number": self.columns_number,
            "rows_number": self.rows_number,
            "info": self.info,
        }

    def list_columns(self) -> list[Dict[str, Any]]:
        """Lists all the columns of the item.

        Returns:
            The columns list.
        """
        return self._client.list_columns(self._id, self._type)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON
from vectice.api.json.model_version import ModelVersionStatus

if TYPE_CHECKING:
    from datetime import datetime

    from vectice.api.json.model_representation import ModelRepresentationOutput


class ModelVersionUpdateInput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def status(self) -> ModelVersionStatus:
        return ModelVersionStatus[str(self["status"])]


class ModelVersionRepresentationOutput(TJSON):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def model_name(self) -> str | None:
        if "model" in self:
            return str(self["model"]["name"])
        return None

    @property
    def status(self) -> str:
        return str(self["status"])

    @property
    def risk(self) -> str:
        return str(self["risk"])

    @property
    def approval(self) -> str:
        return str(self["approval"])

    @property
    def is_starred(self) -> bool:
        return bool(self["isStarred"])

    @property
    def description(self) -> str | None:
        return str(self["description"]) if self["description"] else None

    @property
    def technique(self) -> str | None:
        return str(self["algorithmName"]) if self["algorithmName"] else None

    @property
    def library(self) -> str | None:
        return str(self["framework"]) if self["framework"] else None

    @property
    def inventory_id(self) -> str | None:
        ref = self["inventoryReference"]
        return str(ref) if ref else None

    @property
    def project_id(self) -> str | None:
        if "model" in self:
            return str(self["model"]["project"]["vecticeId"])
        return None

    @property
    def model_id(self) -> str | None:
        if "model" in self:
            return str(self["model"]["vecticeId"])
        return None

    @property
    def phase_origin(self) -> str | None:
        origins = self.get("origins", {})
        phase = origins.get("phase")
        if phase and "vecticeId" in phase:
            return str(phase["vecticeId"])
        return None

    @property
    def iteration_origin(self) -> str | None:
        origins = self.get("origins", {})
        iteration = origins.get("iteration")
        if iteration and "vecticeId" in iteration:
            return str(iteration["vecticeId"])
        return None

    @property
    def creator(self) -> dict[str, str]:
        return {"name": self["createdBy"]["name"], "email": self["createdBy"]["email"]}

    @property
    def model(self) -> ModelRepresentationOutput:
        from vectice.api.json.model_representation import ModelRepresentationOutput

        return ModelRepresentationOutput(**self["model"])

    @property
    def metrics(self) -> List[Dict[str, Any]]:
        return self["metrics"]

    @property
    def properties(self) -> List[Dict[str, Any]]:
        return self["properties"]



from __future__ import annotations

from gql import gql

from vectice.api.gql_api import GqlApi
from vectice.utils.api_utils import PAGINATE_OUTPUT

_RETURNS = PAGINATE_OUTPUT.format(
    """
    key
    value
    __typename
"""
)


class GqlMetricApi(GqlApi):
    def upsert(
        self, type: str, id: str, metrics: list[dict[str, str | float | int]]
    ) -> list[dict[str, str | int | float]]:
        variable_types = (
            "$entityType:EntityMetricType!,$entityId:VecticeId!,$upsertMetricList:[UpsertEntityMetricInput!]!"
        )
        kw = "entityType:$entityType,entityId:$entityId,upsertMetricList:$upsertMetricList"
        variables = {"entityType": type, "entityId": id, "upsertMetricList": metrics}
        gql_query = "upsertEntityMetrics"
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        return self.execute(query_built, variables)[gql_query]["items"]



from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from lightgbm.basic import Booster
    from pyspark.ml.base import Model
    from sklearn.base import BaseEstimator

    from vectice.autolog.model_types import ModelTypes


####### PropertyService for sklearn, xgboost & lightgbm
class PropertyService:

    def _get_model_library(self, model: ModelTypes):
        if "xgboost" in str(model.__class__):
            return "xgboost"
        if "lightgbm" in str(model.__class__):
            return "lightgbm"
        if str(model.__class__.__module__) == "sklearn.model_selection._search":
            return "sklearn-searchcv"
        return "sklearn"

    def _get_lightgbm_info(self, model: Booster) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            params = {
                key: value
                for key, value in model._get_loaded_param().items()  # pyright: ignore[reportPrivateUsage]
                if value is not None
            }
            return "lightgbm", params
        except AttributeError:
            return None, None

    def _get_sklearn_or_xgboost_or_lgbm_info(
        self, model: BaseEstimator | Booster
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        def _flatten_dict(params: dict, parent_key: str = "") -> dict:
            def _reduce_items(acc: list, item: tuple) -> list:
                key, value = item
                new_key = f"{parent_key}_{key}" if parent_key else key
                return (
                    [*acc, *_flatten_dict(value, new_key).items()]
                    if isinstance(value, dict)
                    else [*acc, (new_key, value)]
                )

            return dict(reduce(_reduce_items, params.items(), []))

        try:
            library = self._get_model_library(model)  # pyright: ignore[reportArgumentType]
            base_parameters = (
                model.best_estimator_  # pyright: ignore[reportAttributeAccessIssue]
                if library == "sklearn-searchcv"
                else model
            )
            params = _flatten_dict(
                {
                    str(key): value
                    for key, value in base_parameters.get_params().items()  # pyright: ignore[reportAttributeAccessIssue]
                    if value is not None and bool(str(value))
                }
            )
            return library, params
        except AttributeError:
            return None, None

    def _get_pyspark_parameters(self, model: Model, prefix: str | None = None) -> dict[str, Any]:
        parameters = {}
        for param, value in model.extractParamMap().items():
            if value and isinstance(value, (str, int)) and not isinstance(param, (dict, set)):
                if prefix is not None:
                    parameters[f"{prefix}_{param.name}"] = value
                else:
                    parameters[param.name] = value
        return parameters



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.autolog.asset_services.dataset_service import DatasetService
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.models.resource.metadata.db_metadata import TableType

if TYPE_CHECKING:
    from pyspark.sql import DataFrame as SparkDF
    from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDF


class AutologPysparkService(DatasetService):
    def __init__(
        self,
        key: str,
        asset: SparkDF | SparkConnectDF,
        phase_name: str,
        raw_cells_data: list[str],
        capture_schema_only: bool,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._phase_name = phase_name
        self._prefix = prefix

        super().__init__(raw_cells_data=raw_cells_data, capture_schema_only=capture_schema_only)

    def get_asset(self):
        dataset = self._create_vectice_dataset(self._asset, self._phase_name, self._key, TableType.SPARK, self._prefix)

        return {
            "variable": self._key,
            "dataset": dataset,
            "asset_type": VecticeType.DATASET,
        }



from __future__ import annotations

import ast
import math
import re
from functools import reduce
from typing import TYPE_CHECKING, Any

from vectice import Table
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.utils.code_parser import VariableVisitor, preprocess_code

if TYPE_CHECKING:
    from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]
    from numpy import float_
    from statsmodels.base.wrapper import ResultsWrapper


captured_metrics = set()

# basic metrics
_autolog_metric_allowlist = [
    "aic",
    "bic",
    "centered_tss",
    "condition_number",
    "df_model",
    "df_resid",
    "ess",
    "f_pvalue",
    "fvalue",
    "llf",
    "mse_model",
    "mse_resid",
    "mse_total",
    "rsquared",
    "rsquared_adj",
    "scale",
    "ssr",
    "uncentered_tss",
]


class MetricService:
    def __init__(self, cell_data: dict, custom_metrics: set[str | None] = set()):
        self._cell_data = cell_data
        self._model_cell = None
        self._custom_metrics = custom_metrics

    def _get_metrics(self, cell_content: str) -> list[str]:

        # Get model cell content for metrics
        self._model_cell = preprocess_code(cell_content)

        tree = ast.parse(self._model_cell)
        visitor = VariableVisitor(True, self._custom_metrics)
        visitor.visit(tree)

        return list(visitor.metric_variables)

    def _get_model_metrics(self, data: dict) -> dict[str, Any]:
        cell_content = data["cell"]
        variables = data["variables"]

        if not cell_content:
            return {}

        metric_variables = self._get_metrics(cell_content)

        metrics = reduce(
            lambda identified_metrics, key: (
                {**identified_metrics, key: variables[key]}
                if key in metric_variables and isinstance(variables[key], (int, float))
                else identified_metrics
            ),
            variables.keys(),
            {},
        )
        captured_metrics.update(list(metrics.keys()))
        return metrics

    def get_assets(self):
        metric_tables = self._get_metric_tables(self._cell_data)
        key = "_".join([table.name for table in metric_tables if table is not None])
        return {
            "variable": key,
            "tables": metric_tables,
            "asset_type": VecticeType.METRIC,
        }

    def _get_metric_tables(self, data: dict) -> list[Table | None]:
        cell_content = data["cell"]
        variables = data["variables"]

        if not cell_content:
            return []

        metric_variables = self._get_metrics(cell_content)
        metric_tables = reduce(
            lambda identified_metrics, key: (
                [*identified_metrics, self._parse_metric_table(variables[key], key)]
                if key in metric_variables and isinstance(variables[key], (str, dict)) and key not in captured_metrics
                else identified_metrics
            ),
            variables.keys(),
            [],
        )
        captured_metrics.update([table.name for table in metric_tables if table is not None])
        return metric_tables

    def _parse_metric_table(self, metric_data: str | dict[str, Any], key: str) -> Table | None:
        from pandas import DataFrame

        if isinstance(metric_data, dict):
            return Table(DataFrame(metric_data), key)
        data = self._parse_classification_report(metric_data)
        if data is not None:
            return Table(data, key)
        return None

    def _parse_classification_report(self, report: str):
        from pandas import DataFrame

        lines = report.split("\n")  # Preserve exact structure
        if not lines:
            return None

        # Extract headers
        headers = re.split(r"\s{2,}", lines[0].strip())
        headers.insert(0, "\u00a0")  # Add leading column for labels
        num_columns = len(headers)

        # Process rows
        data = []
        for line in lines[1:]:
            if line.strip():  # Non-empty row
                rows = re.split(r"\s{2,}", line.strip())
                if len(rows) < num_columns:
                    # Rows like 'accuracy' should have placeholders in the beginning
                    rows = [rows[0]] + ["\u00a0"] * (num_columns - len(rows)) + rows[1:]
                data.append(rows)
            else:  # Empty row
                data.append(["\u00a0"] * num_columns)  # Insert a whitespace row

        # Create DataFrame
        df = DataFrame(data, columns=headers)
        return df

    def _get_keras_training_metrics(self, model: KerasModel) -> dict[str, float]:
        try:
            return {
                str(key)
                + "_train": float(model.get_metrics_result()[key].numpy())  # pyright: ignore[reportGeneralTypeIssues]
                for key in model.get_metrics_result().keys()  # pyright: ignore[reportGeneralTypeIssues]
            }
        except Exception:
            return {}

    def _get_statsmodels_metrics(self, model: ResultsWrapper):
        try:
            # statsmodels can function without numpy
            import numpy

            has_numpy = True
        except ImportError:
            has_numpy = False

        def _convert_metric(metric_value: float | float_) -> float | None:
            if metric_value is numpy.float_:  # pyright: ignore[reportPossiblyUnboundVariable]
                metric_value = metric_value.item()

            if math.isnan(float(metric_value)):
                return None
            return round(float(metric_value), 4)

        result_metrics = {}
        for metric in _autolog_metric_allowlist:
            try:
                if hasattr(model, metric):
                    metric_value = getattr(model, metric)
                    if has_numpy:
                        # handle numpy floats
                        converted_metric = _convert_metric(metric_value)
                    else:
                        # simple conversion
                        converted_metric = None if math.isnan(float(metric_value)) else round(float(metric_value), 4)

                    if converted_metric:
                        result_metrics[metric] = _convert_metric(metric_value)
            except Exception:
                pass
        return result_metrics

    def _get_pyspark_ml_summary_metrics(self, summary: Any, prefix: str | None = None) -> dict[str, Any]:
        # metrics
        summary_metrics = {}
        for attr_name in dir(summary):
            # Skip special methods
            if not attr_name.startswith("__"):
                attr_value = getattr(summary, attr_name)
                # get simple values for now, there are pd.DF returns aswell
                if isinstance(attr_value, (float, int)):
                    if prefix:
                        summary_metrics[f"{prefix}_{attr_name}"] = attr_value
                    else:
                        summary_metrics[attr_name] = attr_value
        return summary_metrics



from __future__ import annotations

from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.iteration_asset_representation import IterationAssetRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation

__all__ = [
    "ModelVersionRepresentation",
    "DatasetVersionRepresentation",
    "ModelRepresentation",
    "IterationAssetRepresentation",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from gql import gql

if TYPE_CHECKING:
    from vectice.api.json import ProjectTemplateOutput

from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

_RETURNS = """
            name
            description
            __typename
"""


class ProjectTemplateApi(GqlApi):
    def list_templates(self) -> list[ProjectTemplateOutput]:
        gql_query = "getProjectTemplateList"
        query = GqlApi.build_query(
            gql_query=gql_query,
            returns=_RETURNS,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built)
            output: list[ProjectTemplateOutput] = Parser().parse_list(response[gql_query])
            return output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "workspace", "list")



from __future__ import annotations

from typing import Union

from vectice.api.json.dataset_version import DatasetVersionOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.model_version import ModelVersionOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput

TModelVersion = Union[ModelVersionOutput, ModelVersionRepresentationOutput]
TDatasetVersion = Union[DatasetVersionOutput, DatasetVersionRepresentationOutput]
TVersion = Union[TModelVersion, TDatasetVersion]

ModelVersionClasses = (ModelVersionOutput, ModelVersionRepresentationOutput)
DatasetVersionClasses = (DatasetVersionOutput, DatasetVersionRepresentationOutput)



from __future__ import annotations

import difflib
import inspect
import logging
import os
from functools import reduce
from io import IOBase
from textwrap import dedent
from typing import TYPE_CHECKING, Any, ClassVar

from pandas import DataFrame
from PIL.Image import Image
from rich.table import Table as RichTable

from vectice.api.http_error_handlers import VecticeException
from vectice.api.json.iteration import (
    IterationOutput,
    IterationStatus,
    IterationUpdateInput,
)
from vectice.models.representation.attachment import Attachment
from vectice.models.validation import ValidationModel
from vectice.services.iteration_service import IterationService
from vectice.utils.common_utils import (
    check_read_only,
    ensure_correct_project_id_from_representation_objs,
    temp_print,
    wait_for_path,
)
from vectice.utils.dataframe_utils import repr_list_as_pd_dataframe
from vectice.utils.deprecation import deprecate
from vectice.utils.instance_helper import is_image_or_file
from vectice.utils.last_assets import (
    assign_asset_version_logging,
    comment_or_image_logging,
    register_dataset_logging,
    register_model_logging,
    table_logging,
)
from vectice.utils.logging_utils import get_iteration_status

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Phase, Project, Workspace
    from vectice.models.dataset import Dataset
    from vectice.models.model import Model
    from vectice.models.representation.dataset_representation import DatasetRepresentation
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.iteration_asset_representation import IterationAssetRepresentation
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation
    from vectice.models.table import Table

    TAssetType = ModelRepresentation | ModelVersionRepresentation | DatasetRepresentation | DatasetVersionRepresentation

_logger = logging.getLogger(__name__)


class Iteration:
    """Represents a Vectice iteration.

    An iteration is a recurring work cycle within a phase, primarily used for logging assets (models, datasets, graphs and notes)
    to document your work and maintain transparency throughout an AI project.

    By organizing work into iterations, you can share your key assets, make informed decisions based on previous work, and automatically track and document your progress.

    NOTE: **Each iteration may include sections for organizing assets together, contributing to a more structured narrative to share your work in the way you want.**

    Typical usage example:

    ```python
    # Create a new iteration within a phase
    iteration = phase.create_iteration()

    # Log a dataset to the iteration
    iteration.log(my_dataset)

    # Log a model to the iteration with optional section assignment
    iteration.log(my_model, section="Modeling")
    ```

    NOTE: **You can create sections dynamically using the API.**

    To create a new iteration:

    ```python
    iteration = phase.create_iteration()
    ```
    """

    __slots__: ClassVar[list[str]] = [
        "__dict__",
        "_client",
        "_description",
        "_id",
        "_index",
        "_name",
        "_phase",
        "_service",
        "_status",
        "_steps",
    ]

    def __init__(
        self,
        output: IterationOutput,
        phase: Phase,
        client: Client,
    ):
        self.__dict__ = {}
        self._id = output.id
        self._index = output.index
        self._name = output.name
        self._description = output.description
        self._phase = phase
        self._status = output.status
        self._client = client
        self._service = IterationService(self, self._client)

    def __repr__(self) -> str:
        return f"Iteration (name={self._name}, id={self._id}, status={get_iteration_status(self._status)})"

    def __eq__(self, other: object):
        if not isinstance(other, Iteration):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> str:
        """The iteration's identifier.

        Returns:
            The iteration's identifier.
        """
        return self._id

    @id.setter
    def id(self, iteration_id: str):
        """Set the iteration's identifier.

        Parameters:
            iteration_id: The identifier.
        """
        check_read_only(self)
        self._id = iteration_id

    @property
    def index(self) -> int:
        """The iteration's index.

        Returns:
            The iteration's index.
        """
        return self._index

    @property
    def name(self) -> str:
        """The iteration's name.

        Returns:
            The iteration's name.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def description(self) -> str | None:
        """The iteration's description.

        Returns:
            The iteration's description.
        """
        return self._description

    @description.setter
    def description(self, description: str | None):
        self._description = description

    @property
    def status(self) -> str:
        """The iteration's status.

        Returns:
            The iteration's status.
        """
        return get_iteration_status(self._status)

    @property
    def completed(self) -> bool:
        """Whether this iteration is completed.

        Returns:
            Whether the iteration is completed.
        """
        return self._status is IterationStatus.Completed

    @property
    def properties(self) -> dict[str, int | str]:
        """The iteration's identifier, index and name.

        Returns:
            A dictionary containing the `id`, `index` and `name` items.
        """
        return {"id": self.id, "index": self.index, "name": self.name}

    @deprecate(
        warn_at="24.3.8",
        fail_at="24.4.1",
        remove_at="25.1.1",
        reason="list_phase_requirements is deprecated. Use print_phase_requirements",
    )
    def list_phase_requirements(self) -> None:
        """!!! warning "Deprecation Notice"
        `list_phase_requirements` is deprecated. Use `print_phase_requirements` instead.

        Prints a list of phase requirements in a tabular format, limited to the first 10 requirements. A link is provided to view the remaining requirements.

        Returns:
            None
        """
        paged_requirements = self._client.list_step_definitions(self.phase.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Title", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Description", justify="left", no_wrap=False, min_width=5, max_width=35)

        for requirement in paged_requirements.list:
            description = requirement.description.strip() if requirement.description else None
            rich_table.add_row(requirement.name, description)

        description = f"""There are {paged_requirements.total} requirements in the phase {self.phase.name!r} and a maximum of 10 requirements are displayed in the table below:"""
        link = dedent(
            f"""
                For quick access to the phase requirements in the Vectice web app, visit:
                {self._client.auth.api_base_url}/phase/{self.phase.id}/requirements
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def print_phase_requirements(self) -> None:
        """Prints a list of phase requirements in a tabular format, limited to the first 10 requirements. A link is provided to view the remaining requirements.

        Returns:
            None
        """
        paged_requirements = self._client.list_step_definitions(self.phase.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Title", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Description", justify="left", no_wrap=False, min_width=5, max_width=35)

        for requirement in paged_requirements.list:
            description = requirement.description.strip() if requirement.description else None
            rich_table.add_row(requirement.name, description)

        description = f"""There are {paged_requirements.total} requirements in the phase {self.phase.name!r} and a maximum of 10 requirements are displayed in the table below:"""
        link = dedent(
            f"""
                For quick access to the phase requirements in the Vectice web app, visit:
                {self._client.auth.api_base_url}/phase/{self.phase.id}/requirements
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    @deprecate(
        warn_at="24.3.8",
        fail_at="24.4.1",
        remove_at="25.1.1",
        reason="list_sections is deprecated. Use print_sections instead",
    )
    def list_sections(self) -> None:
        """!!! warning "Deprecation Notice"
        `list_sections` is deprecated. Use `print_sections` instead.

        Prints a list of sections belonging to the iteration in a tabular format, limited to the first 10 sections. A link is provided to view the remaining sections.

        Returns:
            None
        """
        iteration_output = self._client.list_sections(self.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Assets (count)", justify="left", no_wrap=True, min_width=5, max_width=15)

        if iteration_output.artifacts_count > 0:
            rich_table.add_row("None ('Assets without section')", str(iteration_output.artifacts_count))

        for section in iteration_output.sections.list:
            rich_table.add_row(section.name, str(section.artifacts_count))

        description = f"""There are {iteration_output.sections.total} sections in the iteration {self.name!r} and a maximum of 10 sections are displayed in the table below:"""
        link = dedent(
            f"""
                >> To log an asset to a specific section, use iteration.log()
                For quick access to the iteration in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/iteration/{self.id}
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def print_sections(self) -> None:
        """Prints a list of sections belonging to the iteration in a tabular format, limited to the first 10 sections. A link is provided to view the remaining sections.

        Returns:
            None
        """
        iteration_output = self._client.list_sections(self.id)

        rich_table = RichTable(expand=True, show_edge=False)

        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Assets (count)", justify="left", no_wrap=True, min_width=5, max_width=15)

        if iteration_output.artifacts_count > 0:
            rich_table.add_row("None ('Assets without section')", str(iteration_output.artifacts_count))

        for section in iteration_output.sections.list:
            rich_table.add_row(section.name, str(section.artifacts_count))

        description = f"""There are {iteration_output.sections.total} sections in the iteration {self.name!r} and a maximum of 10 sections are displayed in the table below:"""
        link = dedent(
            f"""
                >> To log an asset to a specific section, use iteration.log()
                For quick access to the iteration in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/iteration/{self.id}
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def cancel(self) -> None:
        """Cancel the iteration."""
        iteration_input = IterationUpdateInput(status=IterationStatus.Abandoned.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Abandoned
        _logger.info(f"Iteration {self.name} cancelled.")

    def complete(self) -> None:
        """Mark the iteration as completed."""
        iteration_input = IterationUpdateInput(status=IterationStatus.Completed.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Completed
        logging_output = dedent(
            f"""
                        Iteration {self.name!r} completed.

                        For quick access to the Iteration in the Vectice web app, visit:
                        {self._client.auth.api_base_url}/browse/iteration/{self.id}"""
        ).lstrip()
        _logger.info(logging_output)

    def delete(self) -> None:
        """Permanently deletes the iteration."""
        self._client.delete_iteration(self.id)
        _logger.info(f"Iteration {self.name} was deleted.")

    @property
    def connection(self) -> Connection:
        """The connection to which this iteration belongs.

        Returns:
            The connection to which this iteration belongs.
        """
        return self._phase.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this iteration belongs.

        Returns:
            The workspace to which this iteration belongs.
        """
        return self._phase.workspace

    @property
    def project(self) -> Project:
        """The project to which this iteration belongs.

        Returns:
            The project to which this iteration belongs.
        """
        return self._phase.project

    @property
    def phase(self) -> Phase:
        """The phase to which this iteration belongs.

        Returns:
            The phase to which this iteration belongs.
        """
        return self._phase

    def delete_assets_in_section(self, section: str) -> None:
        """Delete assets within a specified section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_in_section(section="Collect Initial Data")
        ```

        Parameters:
            section: The iteration's section where the assets are displayed in the Vectice App.
        """
        self._client.remove_iteration_assets(iteration_id=self._id, section=section)
        _logger.info(f"All assets in section {section!r} are deleted.")

    def delete_assets_without_a_section(self) -> None:
        """Delete assets without a section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_without_a_section()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id)
        _logger.info("All assets without a section are deleted.")

    def delete_assets_from_iteration(self) -> None:
        """Delete all assets from the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_from_iteration()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id, all=True)
        _logger.info(f"All assets from iteration {self.name!r} are deleted.")

    def delete_asset(self, asset: str) -> None:
        """Delete an asset from the iteration. This action is irreversible, and the deleted assets cannot be recovered.

            asset: The asset Vectice ID, filename or table name.

        ```python
        iteration.delete_asset(asset="MDV-XXXX")
        ```
        """
        asset_id = next(
            (
                artifact._artifact_id  # type: ignore[reportPrivateUsage]
                for artifact in self.list_assets()
                if artifact.asset_id == asset or artifact.asset_name == asset
            ),
            None,
        )
        if asset_id is None:
            raise VecticeException(f"The asset '{asset}' is not valid.")
        self._client.remove_assets_from_iteration(iteration_id=self._id, artifacts_id_list=[int(asset_id)])
        _logger.info(f"Asset '{asset}' from iteration {self.name!r} was deleted.")

    def update(
        self,
        name: str | None = None,
        description: str | None = None,
    ) -> None:
        """Update the Iteration from the API.

            name: The new name of the iteration.
            description: The new description of the iteration.

        Returns:
            None
        """
        self._service.update(name, description)

    def _log_fact_sheet(self, modeva_asset: dict, section: str | None = None):
        self._client.assert_feature_flag_or_raise("modeva-api-integration")  # type: ignore[reportPrivateUsage]

        from vectice import Dataset, DatasetType, NoResource
        from vectice.models.resource.metadata.db_metadata import TableType

        if section is None:
            section = "validation assets"
        dataframe = modeva_asset.get("dataset")
        if dataframe is not None:
            no_resource_dataset = Dataset(
                type=DatasetType.UNKNOWN,
                resource=NoResource(
                    dataframes=dataframe,
                    origin="DATAFRAME",
                    type=TableType.UNKNOWN,
                    capture_schema_only=False,
                ),
            )
            self._log_dataset(no_resource_dataset, section)
        if modeva_asset.get("models"):
            from vectice.models.model import Model

            models = modeva_asset["models"]
            for model in models:
                vect_model = Model(
                    library=model.get("library"),
                    technique=model.get("technique"),
                    properties=model.get("properties"),
                    predictor=model.get("model"),
                )
                self._log_model(vect_model, section)

        if modeva_asset.get("validation_results"):
            from vectice.autolog.autolog_asset_factory import AssetFactory

            validation_results = modeva_asset["validation_results"]
            for validation in validation_results:
                validation_asset = AssetFactory.get_asset_service("none", validation, {}, self.id).get_asset()
                if validation_asset:
                    self._log_validation_result(validation_asset)

    def _get_modeva_docs(self, key: str) -> str | None:
        from modeva import (  # type: ignore[reportMissingImports]
            DataSet,  # type: ignore[reportMissingImports]
            FactSheet,  # type: ignore[reportMissingImports]
        )

        methods_docstrings = {}
        combined_dict = {**DataSet.__dict__, **FactSheet.__dict__}

        for name, member in combined_dict.items():
            try:
                methods_docstrings[name] = inspect.getdoc(member.__func__).split("\n")[0]  # type: ignore[reportOptionalMemberAccess]
            except Exception:
                pass
            try:
                methods_docstrings[name] = inspect.getdoc(member).split("\n")[0]  # type: ignore[reportOptionalMemberAccess]
            except Exception:
                pass

        test_info = methods_docstrings.get(key)
        if test_info is None:
            # closest matching key, fuzzy matching.
            match = difflib.get_close_matches(key, ["eda_1d"], n=1)
            closest_key = match[0] if match else None
            return methods_docstrings.get(closest_key)

        return test_info

    def _log_validation_result(self, modeva_asset: dict, section: str | None = None):
        self._client.assert_feature_flag_or_raise("modeva-api-integration")  # type: ignore[reportPrivateUsage]

        from vectice import Table

        asset = modeva_asset["asset"]

        if section is None:
            section = asset.key

        test_info = self._get_modeva_docs(asset.key)

        if test_info:
            self._log_comment(test_info, section)

        table_prefix = asset.key
        dataframes = modeva_asset.get("table")
        plots = modeva_asset.get("plot")
        if dataframes:
            for index, dataframe in enumerate(dataframes):
                try:
                    # get the multindex as a suffix
                    suffix = dataframe.columns.levels[0][index]
                    # drop the main header level
                    dataframe = dataframe.droplevel(level=0, axis=1)
                    table_name = f"{table_prefix}_{suffix}"
                except Exception:
                    table_name = table_prefix
                table = Table(name=table_name, dataframe=dataframe)
                self._log_table(table, section)
        if plots:
            for plot in plots:
                try:
                    # we need to wait for modeva to save the plot
                    wait_for_path(plot, timeout=20, interval=0.5)
                except TimeoutError as e:
                    print(e)
                self._log_image_or_file(plot, section)

    def _log_scan_report(self, giskard_asset: dict, section: str | None = None):
        self._client.assert_feature_flag_or_raise("giskard-api-integration")

        if section is None:
            section = "Scan Results"

        if giskard_asset.get("results_dataset"):
            dataset = giskard_asset["results_dataset"]
            self._log_dataset(dataset, section)

        if giskard_asset.get("issues"):
            issues = giskard_asset["issues"]
            curr_section = None
            for issue in issues:
                section = issue["section"]
                comment = issue["section_description"]
                level_info = issue["level"]
                table = issue["table"]
                if curr_section is None or (curr_section is not None and curr_section != section):
                    self._log_comment(comment, section)
                    curr_section = section
                self._log_comment(level_info, section)
                self._log_table(table, section)

    def _get_giskard_docs(self, key: str) -> str | None:
        from giskard.rag.report import RAGReport  # type: ignore[reportMissingImports]

        methods_docstrings = {}
        combined_dict = {**RAGReport.__dict__}

        for name, member in combined_dict.items():

            try:
                methods_docstrings[name] = inspect.getdoc(member.__func__).split("\n")[0]  # type: ignore[reportOptionalMemberAccess]
            except Exception:
                pass
            try:
                methods_docstrings[name] = inspect.getdoc(member).split("\n")[0]  # type: ignore[reportOptionalMemberAccess]
            except Exception:
                pass

        test_info = methods_docstrings.get(key)
        return test_info

    def _log_rag_report(self, giskard_asset: dict, section: str | None = None):
        self._client.assert_feature_flag_or_raise("giskard-api-integration")

        if section is None:
            section = "Rag Report"
        if giskard_asset.get("recommendation"):
            recommendation = giskard_asset["recommendation"]
            self._log_comment(recommendation, section)
        if giskard_asset.get("report_dataset"):
            dataset = giskard_asset["report_dataset"]
            self._log_dataset(dataset, section)
        tables = giskard_asset.get("tables")
        if tables is not None:
            for table in tables:
                description = self._get_giskard_docs(table.name)
                if description:
                    self._log_comment(description, section)
                self._log_table(table, section)

    def _log_test_suite_result(self, giskard_asset: dict, section: str | None = None):
        self._client.assert_feature_flag_or_raise("giskard-api-integration")

        attachment = giskard_asset.get("attachment")
        suite_name = giskard_asset.get("suite_name")
        if section is None:
            section = suite_name
        if attachment is not None:
            self._log_image_or_file(attachment, section)

    def _log_qa_test(self, giskard_asset: dict, section: str | None = None):
        from vectice import Dataset, DatasetType, FileResource

        self._client.assert_feature_flag_or_raise("giskard-api-integration")

        if section is None:
            section = "QA Test Set"
        dataframe = giskard_asset.get("dataframe")
        table = giskard_asset.get("table")
        path = giskard_asset.get("csv")
        if dataframe is not None:
            resource = FileResource(path, dataframes=dataframe, capture_schema_only=False)  # type: ignore[reportArgumentType]
            resource_dataset = Dataset(type=DatasetType.UNKNOWN, resource=resource, attachments=[table, path])  # type: ignore[reportArgumentType]
            self._log_dataset(resource_dataset, section)

    def log(self, asset: Any, section: str | None = None) -> None:
        """Log an asset to an iteration. Assets can be organized into sections for improved clarity within the iteration.

        ```python
        from vectice import Model
        my_model = Model(name="my_model")
        iteration.log(my_model)
        iteration.log("my note")
        ```

        Parameters:
            asset: The asset to log to an iteration. Assets can include notes, images, Vectice datasets, or Vectice models.
            section (Optional): The iteration's section where the asset will be displayed in the Vectice App. Assets without sections are logged in the iteration itself.

        """
        # TODO: cyclic imports
        from importlib.util import find_spec

        from vectice.autolog.autolog_asset_factory import AssetFactory
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation
        from vectice.models.table import Table

        is_modeva = find_spec("modeva") is not None
        is_giskard = find_spec("giskard") is not None

        ensure_correct_project_id_from_representation_objs(self.project.id, asset)

        if section:
            section = section.strip()

        if section is not None and len(section) > 60:
            raise VecticeException("section: name must be shorter than or equal to 60 characters")

        giskard_asset = None
        if is_giskard:
            from giskard.rag import QATestset  # type: ignore[reportMissingImports]
            from giskard.rag.report import RAGReport  # type: ignore[reportMissingImports]
            from giskard.scanner.report import ScanReport  # type: ignore[reportMissingImports]

            if isinstance(asset, (QATestset, RAGReport, ScanReport)):
                giskard_asset = AssetFactory.get_asset_service("none", asset, {}, self.id).get_asset()
                if giskard_asset and giskard_asset["type"] == "scan":
                    self._log_scan_report(giskard_asset, section)
                if giskard_asset and giskard_asset["type"] == "rag":
                    self._log_rag_report(giskard_asset, section)
                if giskard_asset and giskard_asset["type"] == "qatest":
                    self._log_qa_test(giskard_asset, section)
                return None

        modeva_asset = None
        if is_modeva:
            from modeva import FactSheet  # type: ignore[reportMissingImports]
            from modeva.utils.results import ValidationResult  # type: ignore[reportMissingImports]

            try:
                if isinstance(asset, (FactSheet, ValidationResult)):
                    modeva_asset = AssetFactory.get_asset_service("none", asset, {}, self.id).get_asset()
                # if modeva_asset and (modeva_asset.get("models") or "dataframe" in modeva_asset):
                #   self._log_fact_sheet(modeva_asset, section)
                if modeva_asset and "table" in modeva_asset:
                    self._log_validation_result(modeva_asset, section)
            except Exception:
                pass

        if modeva_asset or giskard_asset:
            return

        elif is_image_or_file(asset):
            self._log_image_or_file(asset, section)

        elif isinstance(asset, (int, float, str)):
            self._log_comment(asset, section)

        elif isinstance(asset, Table):
            self._log_table(asset, section)

        elif isinstance(
            asset,
            (ModelRepresentation, ModelVersionRepresentation, DatasetRepresentation, DatasetVersionRepresentation),
        ):
            self._assign_version_representation(asset, section)

        elif isinstance(asset, Model):
            if asset.predictor:
                self._assign_model_predictor_metadata(asset)
            self._log_model(asset, section)

        elif isinstance(asset, Dataset):
            self._log_dataset(asset, section)
        elif isinstance(asset, ValidationModel):
            self._log_validation_model(asset, section)
        else:
            raise TypeError(f"Expected Image, Comment, Table, Dataset or a Model, got {type(asset)}")

    def _log_image_or_file(self, asset: str | IOBase | Image, section: str | None = None):
        filename = self._service.log_image_or_file(asset, section)
        comment_or_image_logging(self, _logger, section, filename)

    def _log_comment(self, asset: int | float | str, section: str | None = None):
        self._service.log_comment(asset, section)
        comment_or_image_logging(self, _logger, section)

    def _log_table(self, asset: Table, section: str | None = None):
        self._service.log_table(asset, section)
        table_logging(self, asset, _logger, section)

    def _assign_version_representation(
        self,
        asset: TAssetType,
        section: str | None = None,
    ):
        _, already_assigned = self._service.assign_version_representation(asset, section)
        assign_asset_version_logging(
            iteration=self,
            value=asset,
            _logger=_logger,
            already_assigned=already_assigned,
            section=section,
        )

    def _log_model(self, asset: Model, section: str | None = None):
        model_data, attachments_output, success_pickle = self._service.log_model(asset, section)

        register_model_logging(self, model_data, asset, attachments_output, success_pickle, _logger, section)

    def _log_dataset(self, asset: Dataset, section: str | None = None):
        dataset_output, attachments_output = self._service.log_dataset(asset, section)
        register_dataset_logging(self, dataset_output, asset, attachments_output, _logger, section)

    def _log_validation_model(self, asset: ValidationModel, section: str | None = None):
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        self._client.assert_feature_flag_or_raise("validation-library")

        validated = asset.execute_test()
        metrics = validated.get("metrics") if len(validated.get("metrics")) > 0 else None
        properties = validated.get("properties") if len(validated.get("properties")) > 0 else None
        attachments = validated.get("attachments") if len(validated.get("attachments")) > 0 else None

        if asset.asset:
            mdv = ModelVersionRepresentation(output=self._client.get_model_version(asset.asset), client=self._client)
            mdv.update(metrics=metrics, properties=properties, attachments=attachments)
            self._assign_version_representation(mdv)
        else:
            if attachments is not None:
                for attachment in attachments:
                    self._log_image_or_file(attachment, section)

        tables = validated.get("table")
        if tables is not None:
            for table in tables:
                self._log_table(table, section)

    def _assign_model_predictor_metadata(self, asset: Model) -> None:
        from vectice.autolog.autolog_asset_factory import AssetFactory

        asset_information = AssetFactory.get_asset_service(
            "key", asset.predictor, {"cell": None, "variables": None}, self.id
        ).get_asset()
        if asset_information and asset_information["model"]:
            model = asset_information["model"]
            technique, params, library = (
                model.technique,
                model.properties,
                model.library,
            )
            if technique and asset.technique is None:
                asset.technique = technique
            if params and asset.properties is None:
                # disable update model properties warning
                logger = logging.getLogger("vectice.models.model")
                logger.disabled = True

                asset.properties = params

                logger.disabled = False
            if library and asset.library is None:
                asset.library = library

    def list_assets(self) -> list[IterationAssetRepresentation]:
        """Retrieves a list of assets associated with the iteration.

        Returns:
            list[IterationAssetRepresentation]: A list of IterationAssetRepresentation objects representing the assets
                associated with the iteration.
        """
        from vectice.models.representation.iteration_asset_representation import IterationAssetRepresentation

        return reduce(
            lambda acc, asset: [*acc, IterationAssetRepresentation(asset, self._client)],
            self._client.list_iteration_assets(self.id).list,
            [],
        )

    def list_attachments(self) -> list[Attachment]:
        """Lists all the attachments (tables, files, pickle...) of the Iteration.

        Returns:
            The attachments list.
        """
        return self._client.list_attachments(self)

    def download_attachments(self, attachments: list[str] | str | None = None, output_dir: str | None = None) -> None:
        """Downloads a list of attachments associated with the current iteration.

        Parameters:
            attachments: A list of attachment file names or a single attachment file name
                                                  to be downloaded. If None, all attachments will be downloaded.
            output_dir: The directory path where the attachments will be saved.
                                      If None, the current working directory is used.

        Returns:
            None
        """
        if output_dir is None:
            output_dir = os.getcwd()
        os.makedirs(output_dir, exist_ok=True)

        if attachments is None:
            attachments = list(map(lambda attach: attach.name, self.list_attachments()))

        if isinstance(attachments, str):
            attachments = [attachments]

        for attachment in attachments:
            self._client.download_attachment(self, attachment, output_dir)

    def get_table(self, table: str) -> DataFrame:
        """Retrieves a table associated with the current iteration.

        Parameters:
            table: The name of the table.

        Returns:
            The data from the specified table as a DataFrame.
        """
        return repr_list_as_pd_dataframe(self._client.get_version_table(self, table))



from __future__ import annotations

from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.review import ReviewOutput
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json.paged_response import PagedResponse


_RETURNS = """vecticeId
              status
              feedback
              message
              createdDate
              reviewer {
                name
                email
              }
              createdBy {
                name
                email
              }
              __typename
            """


_RETURNS_PAGE = PAGINATE_OUTPUT.format(_RETURNS)


class GqlReviewApi(GqlApi):
    def list_reviews(self, parent_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS) -> PagedResponse[ReviewOutput]:
        gql_query = "getReviewList"

        variable_types = "$filters:ReviewFiltersInput!,$order:ListOrderInput,$page:PageInput"
        kw = "filters:$filters,order:$order,page:$page"
        variables = {
            "filters": {
                "phaseId": parent_id,
            },
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS_PAGE, keyword_arguments=kw
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: PagedResponse[ReviewOutput] = Parser().parse_paged_response(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "phase", "list")



from __future__ import annotations

import logging
from typing import Any, Dict

from pandas.core.frame import DataFrame as PandasDF

from vectice.models.resource.metadata.base import MetadataSettings
from vectice.models.resource.metadata.column_metadata import Column
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.df_wrapper_pandas import PandasDFWrapper
from vectice.models.resource.metadata.df_wrapper_pyspark_pandas import PysparkPandasDFWrapper
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper
from vectice.models.resource.metadata.df_wrapper_spark_df import SparkDFWrapper
from vectice.models.resource.metadata.extra_metadata import ExtraMetadata

_logger = logging.getLogger(__name__)

TDataFrameType = DataFrameType


class Source:
    def __init__(
        self,
        name: str,
        size: int | None = None,
        columns: list[Column] | None = None,
        updated_date: str | None = None,
        created_date: str | None = None,
        uri: str | None = None,
        dataframe: TDataFrameType | None = None,
        extra_metadata: list[ExtraMetadata] | None = None,
        display_name: str | None = None,
        capture_schema_only: bool = False,
    ):
        """Initialize a MetadataDB instance.

        Parameters:
            name: The name of the source.
            size: The size of the source.
            columns: The columns that compose the source.
            updated_date: The date of last update of the source.
            created_date: The date of creation of the source.
            uri: The uri of the source.
            dataframe (Optional): A dataframe allowing vectice to optionally compute more metadata about this resource such as columns stats, size, rows number and column numbers. (Support Pandas and Spark)
            extra_metadata (Optional): Extra metadata to be captured.
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.

        """
        self.name = name
        self.display_name = display_name
        self.size = size
        self.columns = columns
        self.created_date = created_date
        self.updated_date = updated_date
        self.uri = uri
        self.extra_metadata = extra_metadata
        self._dataframe = dataframe
        self._wrapper: PandasDFWrapper | PysparkPandasDFWrapper | SparkDFWrapper | None = None
        self._settings = MetadataSettings()
        self.capture_schema_only = capture_schema_only

        if isinstance(self._dataframe, PandasDF):
            self._wrapper = PandasDFWrapper(self._dataframe)
        elif isinstance(self._dataframe, DataFrameWrapper):
            _logger.warning(
                "WARNING: Custom wrappers are not supported yet, pass a Pandas or a Spark DataFrame to get statistics."
            )
        else:
            try:
                from pyspark.pandas.frame import DataFrame as PysparkPandasDF
                from pyspark.sql.dataframe import DataFrame as SparkDF

                if isinstance(self._dataframe, SparkDF):
                    self._wrapper = SparkDFWrapper(self._dataframe)
                elif isinstance(self._dataframe, PysparkPandasDF):
                    _logger.warning(
                        "WARNING: Pandas on spark DataFrame is not supported yet, pass a Pandas or a Spark DataFrame to get statistics."
                    )
            except ImportError:
                pass

    def set_settings(self, settings: MetadataSettings):
        self._settings = settings

    def asdict(self) -> dict[str, Any]:
        size_info: Dict[str, int | float | None] = {}
        skipped_statistics = False
        minimum_rows_for_statistics = self._settings.minimum_rows_for_statistics
        sample_rows_for_statistics = self._settings.sample_rows_for_statistics
        maximum_columns_for_statistics = self._settings.maximum_columns_for_statistics
        if self._wrapper is not None:
            df_info = self._wrapper.get_size()
            if df_info.rows < minimum_rows_for_statistics:
                skipped_statistics = True
                _logger.warning(
                    f"Statistics are not captured if numbers of rows are below {minimum_rows_for_statistics}, to keep the data anonymous."
                )
            size_info = df_info.asdict()

        columns_list: list[Column] = self.columns if self.columns is not None else []
        columns_list = (
            self._wrapper.capture_columns(
                minimum_rows_for_statistics,
                sample_rows_for_statistics,
                maximum_columns_for_statistics,
                capture_schema_only=self.capture_schema_only,
            )
            if self._wrapper is not None
            else columns_list
        )
        columns_list_dict = [col.asdict() for col in columns_list]
        exm = self.extra_metadata
        return {
            **size_info,
            "size": self.size,
            "name": self.name,
            "updatedDate": self.updated_date,
            "createdDate": self.created_date,
            "uri": self.uri,
            "columns": columns_list_dict,
            "skippedStatistics": skipped_statistics,
            "extraMetadata": [metadata.to_dict() for metadata in exm] if exm is not None else None,  # type: ignore[attr-defined]
        }



from __future__ import annotations

from typing import Any

from vectice.autolog.asset_services.service_types.modeva_types import ModevaType


class ModevaValidationResult:
    def __init__(self, key: str, asset: Any, data: dict):
        self._asset = asset
        self._key = key

    def get_asset(self):
        try:
            plot = self._get_plot(self._asset)
            table = self._get_table()
            return {
                "variable": self._key,
                "table": table,
                "plot": plot,
                "asset": self._asset,
                "asset_type": ModevaType.VALIDATION_RESULT,
            }
        except Exception:
            pass

    def _normalize_dataframe(self, dataframe: Any) -> Any:
        # normalize dataframe so FE can parse the df
        # this is for compare tests aswell
        if "diagnose_robustness" == self._asset.key:
            return dataframe.reset_index().melt(id_vars=["Repeats"], value_name="Value")
        if "reliability" in self._asset.key:
            # FE can't parse col name with a .
            try:
                dataframe.columns = dataframe.columns.str.replace("[^A-Za-z0-9_]+", "_", regex=True)
            except Exception:
                pass
            return dataframe
        return dataframe

    def _get_table(self) -> list:
        df = self._asset.table if hasattr(self._asset, "table") else None
        if df is None:
            return []
        multi_index = df.columns.nlevels > 1
        if not multi_index:
            return [self._normalize_dataframe(df)]
        dataframes = []
        # split multi-index into separate dfs
        for _, group in df.groupby(level=0, axis=1):
            dataframes.append(self._normalize_dataframe(group))
        return dataframes

    def _get_plot(self, result: Any) -> list[str]:
        # no plots
        if result.options is None:
            return []
        plot_suffixes = result.options.keys()
        # chart_id only appears in single plots
        multiple_plots = len(plot_suffixes) > 1 and "chart_id" not in plot_suffixes
        full_path = f"vectice_modeva/{result.key}"
        # temp_file_path = rf"{temp_dir.name}\{file_name}"
        result.plot_save(file_name=full_path, format="png", figsize=(6, 4))
        if multiple_plots:
            all_plot_paths = [f"{full_path}-{suffix}.png" for suffix in plot_suffixes]
            # full path will be a prefix
            return all_plot_paths
        return [full_path + ".png"]



from __future__ import annotations

import inspect
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from typing import Any


class PropertyInput:
    def __init__(self, key: str, value: str | int | float, timestamp: datetime | str | None = None):
        self.key: str = key
        self.value: str | int | float = str(value)
        if timestamp is None:
            self.timestamp: str = datetime.now(timezone.utc).isoformat()
        else:
            self.timestamp = timestamp.isoformat() if isinstance(timestamp, datetime) else str(timestamp)


@dataclass
class PropertyOutput:
    id: int
    dataSetVersionId: int
    key: str
    value: str
    timestamp: datetime
    name: str | None = None

    @classmethod
    def from_dict(cls, properties: Any):
        return cls(**{k: v for k, v in properties.items() if k in inspect.signature(cls).parameters})

    def as_dict(self):
        return asdict(self)


def create_properties_input(properties: dict[str, int | float | str]) -> list[PropertyInput]:
    if len(set(properties)) < len(properties):
        raise ValueError("You can not use the same key value pair more than once.")
    props: list[PropertyInput] = []
    for key, value in properties.items():
        _check_empty_property(key, value)
        props.append(PropertyInput(key, str(value)))
    return props


def _check_empty_property(key: str, value: str | int | float):
    if key.strip() == "" or (isinstance(value, str) and value.strip() == ""):
        raise ValueError("Property keys and values can't be empty.")



from __future__ import annotations

import datetime
import re
from typing import Any, Optional

from dateutil import parser

try:
    from pyspark.sql import SparkSession
    from pyspark.sql.connect.session import SparkSession as SparkConnectSession

    spark_imported = True
except ImportError:
    spark_imported = False
    pass

from vectice.models.resource.base import Resource
from vectice.models.resource.databricks import DatabricksTableSparkConnectSessionSQL, DatabricksTableSparkSessionSQL
from vectice.models.resource.metadata.column_metadata import DBColumn
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.db_metadata import DBMetadata, MetadataDB
from vectice.models.resource.metadata.extra_metadata import ExtraMetadata

TDataFrameType = DataFrameType


class DatabricksTableResource(Resource):
    """Databricks tables resource reference wrapper.

    This resource wraps Databricks tables paths that you have stored in Databricks with optional metadata and versioning.
    You pass it as an argument of your Vectice Dataset wrapper before logging it to an iteration.


    ```python
    from vectice import DatabricksTableResource

    db_resource = DatabricksTableResource(
        spark_client=spark,
        paths="my_table",
    )
    ```

    NOTE: **Vectice does not store your actual dataset. Instead, it stores your dataset's metadata, which is information about your dataset.
    These details are captured by resources tailored to the specific environment in use, such as: local (FileResource), Bigquery, S3, GCS...**


    Known limitations :
        - Resource size won't be captured when passing a specific version. Example: my_table@v2.
    """

    _origin = "DATABRICKS_TABLE"

    def __init__(
        self,
        paths: str | list[str],
        dataframes: TDataFrameType | list[TDataFrameType] | None = None,
        spark_client: SparkSession | None = None,
        capture_schema_only: bool = False,
        columns_description: dict[str, str] | str | None = None,
    ):
        """Initialize a DatabricksTableResource resource.

        Parameters:
            paths: The paths to retrieve the tables. Should be either the table name, the location of the table or full path of the table for Spark Connect.
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas, Spark)
            spark_client (Optional): The spark session allowing vectice to capture the table metadata.
            capture_schema_only (Optional): A boolean parameter indicating whether to capture only the schema or both the schema and column statistics of the dataframes.
            columns_description (Optional): A dictionary or path to a csv file to map the column's name to a specific description. Dictionary should follow the format { "column_name": "Description", ... }

        """
        if spark_imported is False:
            raise ImportError("Pyspark is not installed.")
        super().__init__(
            paths=paths,
            dataframes=dataframes,
            capture_schema_only=capture_schema_only,
            columns_description=columns_description,
        )
        self._spark_client = spark_client
        if spark_client and isinstance(
            spark_client, SparkConnectSession  # pyright: ignore[reportPossiblyUnboundVariable]
        ):
            self._databricks_table_sql = DatabricksTableSparkConnectSessionSQL(spark_client=spark_client)
        elif spark_client is not None:
            self._databricks_table_sql = DatabricksTableSparkSessionSQL(spark_client=spark_client)
        else:
            self._databricks_table_sql = None

    def _fetch_data(self):
        tables: dict[str, tuple[dict | None, TDataFrameType | None]] = {}
        df_index = 0
        for path in self._paths:
            path_tables, new_df_index = self._fetch_table_data(index=df_index, path=path)
            tables.update(path_tables)
            df_index += new_df_index
        return tables

    def _fetch_table_data(
        self, index: int, path: str
    ) -> tuple[dict[str, tuple[dict | None, TDataFrameType | None]], int]:
        table_display_name = path
        has_at = re.search("@", path)
        has_version = re.search("@v", path)
        table_path = path.split("@")[0] if has_at else path

        dataframe = self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None

        if self._spark_client is None:
            return {table_display_name: (None, dataframe)}, 1

        if self._databricks_table_sql is None:
            raise ValueError("A spark_client of type `SparkSession` is required.")
        self._databricks_table_sql.table_path = table_path

        table_name = table_path.rpartition("/")[-1]
        self._databricks_table_sql.set_table_key(table_name)
        schema: list | None = None
        detail: dict = {
            "location": None,
            "createdAt": None,
            "format": None,
            "lastModified": None,
            "sizeInBytes": None,
        }

        location: str | None = None
        try:
            detail = self._databricks_table_sql.get_detail_as_dict()
            if has_at is not None or has_version is not None:
                detail["sizeInBytes"] = None

            location = detail["location"]
            if location is not None:
                table_path = location
        except Exception:  # Handle views
            rows = self._databricks_table_sql.get_rows()

            row_created = next(row for row in rows if row["col_name"] == "Created Time")
            detail["createdAt"] = row_created["data_type"]
            end_row = next(row for row in rows if row["col_name"] == "" and row["data_type"] == "")
            schema = rows[: rows.index(end_row)]

        if schema is None:
            schema = self._get_schema(self._spark_client, has_version, has_at, path, table_path)

        table_format = detail["format"]
        extra_metadata: list[ExtraMetadata] = [
            ExtraMetadata(key="format", value=table_format, display_name="Format"),
        ]

        information = self._get_table_information()
        if "Type" in information:
            extra_metadata.append(ExtraMetadata(key="type", value=information["Type"], display_name="Type"))
        if detail["sizeInBytes"] is None and "Statistics" in information and has_at is None and has_version is None:
            match_bytes = re.search(r"([0-9]{1,15})( bytes)", information["Statistics"])
            if match_bytes is not None:
                detail["sizeInBytes"] = int(match_bytes.group(1))

        if table_format == "delta":
            extra_metadata.extend(self._get_delta_table_history_extra_metadata(has_version, has_at, path))

        if dataframe is None:
            key_rows_table_name = path.rpartition("/")[-1]
            self._databricks_table_sql.set_rows_table_name(key_rows_table_name)
            detail["rowsNumber"] = self._databricks_table_sql.get_number_of_rows()

        return {
            table_display_name: ({"detail": detail, "schema": schema, "extra_metadata": extra_metadata}, dataframe)
        }, 1

    def _get_schema(
        self,
        spark_session: SparkSession,
        has_version: Optional[re.Match[str]],
        has_at: Optional[re.Match[str]],
        path: str,
        table_path: str,
    ):
        if has_version:
            table = spark_session.read.option("versionAsOf", path.split("@v")[1]).load(table_path)
        elif has_at:
            table = spark_session.read.option("timestampAsOf", path.split("@")[1]).load(table_path)
        else:
            if self._databricks_table_sql is None:
                raise ValueError("A spark_client of type `SparkSession` is required.")
            try:
                table = self._databricks_table_sql.read_table_path(table_path)
            except Exception:
                return self._databricks_table_sql.get_detail()

        return list(map(lambda row: {"col_name": row.name, "data_type": row.dataType.jsonValue()}, table.schema))

    def _get_table_information(self) -> dict:
        if self._databricks_table_sql is None:
            raise ValueError("A spark_client of type `SparkSession` is required.")
        df = self._databricks_table_sql.get_table_extended()
        df_collected = df.collect()
        if len(df_collected) == 0:
            return {}
        dict_df_table = df_collected[0].asDict()
        if "information" not in dict_df_table:
            return {}
        information: str = dict_df_table["information"]
        pairs = information.split("\n")
        result_dict = {}

        for pair in pairs:
            if ": " in pair:
                key, value = pair.split(": ")
                key = key.strip()
                value = value.strip()
                result_dict[key] = value

        return result_dict

    def _get_delta_table_history_extra_metadata(
        self,
        has_version: Optional[re.Match[str]],
        has_at: Optional[re.Match[str]],
        path: str,
    ) -> list[ExtraMetadata]:
        if self._databricks_table_sql is None:
            raise ValueError("A spark_client of type `SparkSession` is required.")
        history = self._databricks_table_sql.get_table_history()
        if has_version:
            history_v = path.split("@v")[1]
            history_row = next(row for row in history if row["version"] == int(history_v))
        elif has_at:
            history_v = path.split("@")[1]
            history_row = next(row for row in history if row["timestamp"] == parser.parse(history_v))
        else:
            history_row = history[0]

        version = history_row["version"]
        date_time = history_row["timestamp"]
        return [
            ExtraMetadata(key="version", value=f"v{version}", display_name="Version"),
            ExtraMetadata(key="timestamp", value=str(date_time.isoformat()), display_name="Timestamp"),
        ]

    def _build_metadata(self) -> DBMetadata:
        tables_metadata: dict[str, tuple[dict | None, TDataFrameType | None]] = self.data
        dbs: list[MetadataDB] = []
        for table_name, [table_metadata, dataframe] in tables_metadata.items():
            columns: list[DBColumn] | None = None
            size = None
            updated_date = None
            created_date = None
            rows_number = None
            uri = None
            if table_metadata is not None and self._spark_client is not None:
                columns = []
                detail = table_metadata["detail"]
                created_at = detail["createdAt"]
                last_modified = detail["lastModified"]
                size = detail["sizeInBytes"]
                uri = detail["location"]
                if "rowsNumber" in detail:
                    rows_number = detail["rowsNumber"]

                def get_iso_format_from_str_or_date(date_time: Any) -> str | None:
                    if date_time is not None:
                        if isinstance(date_time, str):
                            return parser.parse(date_time).isoformat()
                        elif isinstance(date_time, datetime.datetime):
                            return date_time.isoformat()
                    return None

                created_date = get_iso_format_from_str_or_date(created_at)
                updated_date = get_iso_format_from_str_or_date(last_modified)

                for row in table_metadata["schema"]:
                    columns.append(
                        DBColumn(
                            name=row["col_name"],
                            data_type=row["data_type"],
                        )
                    )

            dbs.append(
                MetadataDB(
                    name=table_name,
                    size=size,
                    rows_number=rows_number,
                    columns=columns,
                    created_date=created_date,
                    updated_date=updated_date,
                    uri=uri,
                    dataframe=dataframe,
                    extra_metadata=table_metadata["extra_metadata"] if table_metadata is not None else None,
                    display_name=table_name.rpartition("/")[-1],
                    capture_schema_only=self.capture_schema_only,
                )
            )

        metadata_size = None
        for db in dbs:
            if db.size is not None:
                if metadata_size is None:
                    metadata_size = 0
                metadata_size += db.size

        metadata = DBMetadata(
            size=metadata_size,
            origin=self._origin,
            dbs=dbs,
        )
        return metadata



from __future__ import annotations

import logging
import os
from enum import Enum
from json import JSONEncoder
from typing import TYPE_CHECKING, Any, BinaryIO, Dict, Sequence, cast

import requests

from vectice.api.http_error import HttpError
from vectice.api.http_error_handlers import ClientErrorHandler

if TYPE_CHECKING:
    from requests import Response

    from vectice.api._auth import Auth
    from vectice.api.json_object import JsonObject

logger = logging.getLogger("vectice.api.rest")


def log_request(method: str, path: str, headers: dict[str, str], payload: Any | None = None) -> None:
    should_log = os.getenv("LOG_VECTICE_HTTP_REQUEST")
    if should_log is not None and should_log != "0" and should_log.lower() != "false":
        logger.info("###")
        logger.info(f"{method} {path}")
        for item in headers.items():
            logger.info(f"{item[0]}: {item[1] if item[0] != 'Authorization' else '********'}")
        logger.info(payload)
    else:
        logger.debug("###")
        logger.debug(f"{method} {path}")
        for item in headers.items():
            logger.debug(f"{item[0]}: {item[1] if item[0] != 'Authorization' else '********'}")
        logger.debug(payload)


class VecticeEncoder(JSONEncoder):
    """JSON encoder with 2 specific behaviors.

    - handle datetime types so be serialized as a string following ISO8601 format
    - remove any null property from the serialized JSON
    - handle nested objects
    """

    def default(self, obj: Any) -> Any:  # pyright: ignore[reportIncompatibleMethodOverride]
        from copy import deepcopy

        if hasattr(obj, "isoformat"):
            return obj.isoformat()
        if isinstance(obj, Enum):
            return obj.value
        internal_copy = deepcopy(obj.__dict__)
        return {k.lstrip("_"): v for (k, v) in internal_copy.items() if v is not None}


class RestApi:
    def __init__(self, auth: Auth):
        self.auth = auth
        self.api_base_url = self.auth.api_base_url
        self.verify_certificate = self.auth.verify_certificate
        self._httpErrorHandler = ClientErrorHandler()

    def get(self, path: str) -> dict[str, Any]:
        headers = self._add_json_response_to_headers()
        log_request("GET", path, headers)
        response = requests.get(  # noqa: S113
            url=self.api_base_url + path, headers=headers, verify=self.verify_certificate
        )
        return self._response(self.api_base_url + path, headers, response, "GET")

    def post(self, path: str, payload: Any = None) -> JsonObject:
        headers = self._add_json_response_to_headers()
        cleaned_payload: Any = self._clean_dict(payload) if isinstance(payload, dict) else payload
        data = VecticeEncoder(indent=1).encode(cleaned_payload)
        log_request("POST", path, headers, data)
        response = requests.post(  # noqa: S113
            url=self.api_base_url + path, headers=headers, data=data, verify=self.verify_certificate
        )
        return self._response(self.api_base_url + path, headers, response, "POST", data)

    def put(self, path: str, payload: Any = None) -> JsonObject:
        headers = self._add_json_response_to_headers()
        cleaned_payload: Any = self._clean_dict(payload) if isinstance(payload, dict) else payload
        data = VecticeEncoder(indent=1).encode(cleaned_payload)
        log_request("PUT", path, headers, data)
        response = requests.put(  # noqa: S113
            url=self.auth.api_base_url + path, headers=headers, data=data, verify=self.verify_certificate
        )
        return self._response(self.auth.api_base_url + path, headers, response, "PUT", payload)

    def delete(self, path: str, payload: Any = None) -> JsonObject:
        headers = self._add_json_response_to_headers()
        if payload is None:
            data = None
        else:
            cleaned_payload: Any = self._clean_dict(payload) if isinstance(payload, dict) else payload
            data = VecticeEncoder(indent=1).encode(cleaned_payload)
        log_request("DELETE", path, headers, data)
        response = requests.delete(  # noqa: S113
            url=self.auth.api_base_url + path, headers=headers, data=data, verify=self.auth.verify_certificate
        )
        return self._response(self.auth.api_base_url + path, headers, response, "DELETE", data)

    def _post_attachments(
        self, path: str, files: Sequence[tuple[str, tuple[Any, BinaryIO | str]]] | None = None
    ) -> Response | None:
        from vectice import auto_extract

        headers = self.auth.http_headers
        params = {"extract": str(auto_extract).lower()}
        response = requests.post(  # noqa: S113
            url=self.auth.api_base_url + path,
            headers=headers,
            files=files,
            verify=self.auth.verify_certificate,
            params=params,
        )
        return self._attachment_response(self.auth.api_base_url + path, headers, response, "POST")

    def _put_attachments(
        self, path: str, files: list[tuple[str, tuple[Any, BinaryIO]]] | None = None
    ) -> Response | None:
        headers = self.auth.http_headers
        response = requests.put(  # noqa: S113
            url=self.api_base_url + path,
            headers=headers,
            files=files,
            verify=self.verify_certificate,
        )
        return self._attachment_response(self.api_base_url + path, headers, response, "PUT")

    def _get_attachment(self, path: str) -> Response:
        headers = self.auth.http_headers
        response = requests.get(  # noqa: S113
            url=self.api_base_url + path, headers=headers, verify=self.verify_certificate, stream=True
        )
        return self._attachment_response(self.api_base_url + path, headers, response, "GET")

    def _delete_attachment(self, path: str) -> Response | None:
        headers = self.auth.http_headers
        response = requests.delete(url=self.api_base_url + path, headers=headers)  # noqa: S113
        return self._attachment_response(self.api_base_url + path, headers, response, "DELETE")

    def _list_attachments(self, path: str) -> list[dict[str, Any]]:
        headers = self._add_json_response_to_headers()
        response = requests.get(  # noqa: S113
            url=self.api_base_url + path, headers=headers, verify=self.verify_certificate
        )
        self._attachment_response(self.api_base_url + path, headers, response, "GET")
        return response.json()

    def _add_json_response_to_headers(self) -> dict[str, str]:
        return {**self.auth.http_headers, "Content-Type": "application/json"}

    @classmethod
    def raise_status(cls, path: str, response: Response, method: str, payload: Any | None = None) -> None:
        if not (200 <= response.status_code < 300):
            reason = response.text
            if not isinstance(payload, str):
                json = VecticeEncoder(indent=4, sort_keys=True).encode(payload) if payload is not None else None
            else:
                json = payload
            raise HttpError(response.status_code, reason, path, method, json)

    def _response(
        self, path: str, headers: dict[str, str], response: Response, method: str, payload: Any | None = None
    ) -> JsonObject:
        self.raise_status(path, response, method, payload)
        logger.debug(f"{method} {path} {response.status_code}")
        logger.debug("\n".join(f"{item[0]}: {item[1]}" for item in headers.items()))
        logger.debug(payload)
        if len(response.content) > 0:
            return cast(Dict[str, Any], response.json())
        return {response.reason: response.status_code}

    def _attachment_response(self, path: str, headers: dict[str, str], response: Response, method: str) -> Response:
        self.raise_status(path, response, method)
        logger.debug(f"{method} {path} {response.status_code}")
        logger.debug("\n".join(f"{item[0]}: {item[1]}" for item in headers.items()))
        return response

    def _clean_dict(self, payload: dict[Any, Any]):
        cleaned_payload = {}
        for key, value in payload.items():
            if value is not None:
                if isinstance(value, dict):
                    cleaned_payload[key] = self._clean_dict(value)
                else:
                    cleaned_payload[key] = value
        return cleaned_payload



from __future__ import annotations

from vectice.api.json.requirement import RequirementOutput


class SectionOutput(RequirementOutput):
    @property
    def artifacts_count(self) -> int:
        return int(self["paginatedArtifacts"]["total"])



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api.json.json_type import TJSON
from vectice.api.json.metric import MetricInput

if TYPE_CHECKING:
    from vectice.api.json.model_version import ModelVersionOutput, ModelVersionStatus
    from vectice.models.property import Property


class ModelRegisterOutput(TJSON):
    @property
    def model_version(self) -> ModelVersionOutput:
        from vectice.api.json.model_version import ModelVersionOutput

        return ModelVersionOutput(**self["modelVersion"])

    @property
    def use_existing_model(self) -> bool:
        return bool(self["useExistingModel"])

    # TODO: complete the jobrun property
    @property
    def job_run(self) -> str:
        return str(self["jobRun"])


class ModelRegisterInput(TJSON):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def model_type(self) -> str:
        return str(self["modelType"])

    @property
    def properties(self) -> Property:
        from vectice.models.property import Property

        return Property(**self["properties"])

    @property
    def metrics(self) -> list[MetricInput]:
        return [MetricInput(**metric) for metric in self["metrics"]]

    @property
    def status(self) -> ModelVersionStatus:
        from vectice.api.json.model_version import ModelVersionStatus

        return ModelVersionStatus(self["status"])

    @property
    def framework(self) -> str | None:
        return str(self["framework"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def algorithm_name(self) -> str | None:
        return str(self["algorithmName"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def dataset_inputs(self) -> list[str] | None:
        return list(self["datasetInputs"]) if self["datasetInputs"] else None

    @property
    def model_inputs(self) -> list[str] | None:
        return list(self["modelInputs"]) if self["modelInputs"] else None

    @property
    def context(self) -> ModelVersionContextInput:
        return ModelVersionContextInput(self["context"])


class ModelVersionContextInput(TJSON):
    @property
    def url(self) -> str:
        return str(self["url"])

    @property
    def run(self) -> str:
        return str(self["run"])



from __future__ import annotations

from datetime import datetime
from typing import Any

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput


class ReportOutput(TJSON):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    @property
    def id(self) -> int:
        return self["id"]

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def model_version(self) -> ModelVersionRepresentationOutput | None:
        if self["modelVersion"] is not None:
            return ModelVersionRepresentationOutput(self["modelVersion"])

        return None

    @property
    def creator(self) -> dict[str, str]:
        return {"name": self["createdBy"]["name"], "email": self["createdBy"]["email"]}
